from copy import deepcopy

import mock
import pytest
import requests
import requests_mock

from pubtools._quay import manifest_list_merger, quay_client

old_ml = {
    "schemaVersion": 2,
    "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
    "manifests": [
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:1111111111",
            "platform": {"architecture": "arm64", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:2222222222",
            "platform": {"architecture": "armhfp", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:3333333333",
            "platform": {"architecture": "ppc64le", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:5555555555",
            "platform": {"architecture": "amd64", "os": "linux"},
        },
    ],
}

new_ml = {
    "schemaVersion": 2,
    "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
    "manifests": [
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:6666666666",
            "platform": {"architecture": "arm64", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:7777777777",
            "platform": {"architecture": "ppc64le", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:8888888888",
            "platform": {"architecture": "s390x", "os": "linux"},
        },
    ],
}

merged_ml = {
    "schemaVersion": 2,
    "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
    "manifests": [
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:6666666666",
            "platform": {"architecture": "arm64", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:2222222222",
            "platform": {"architecture": "armhfp", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:7777777777",
            "platform": {"architecture": "ppc64le", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:8888888888",
            "platform": {"architecture": "s390x", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:5555555555",
            "platform": {"architecture": "amd64", "os": "linux"},
        },
    ],
}

merged_ml2 = {
    "schemaVersion": 2,
    "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
    "manifests": [
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:6666666666",
            "platform": {"architecture": "arm64", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:2222222222",
            "platform": {"architecture": "armhfp", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:3333333333",
            "platform": {"architecture": "ppc64le", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:5555555555",
            "platform": {"architecture": "amd64", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:8888888888",
            "platform": {"architecture": "s390x", "os": "linux"},
        },
    ],
}

merged_ml3 = {
    "schemaVersion": 2,
    "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
    "manifests": [
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:6666666666",
            "platform": {"architecture": "arm64", "os": "linux"},
        },
        {
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "size": 429,
            "digest": "sha256:8888888888",
            "platform": {"architecture": "s390x", "os": "linux"},
        },
    ],
}

expected_missing_archs = [
    {
        "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
        "size": 429,
        "digest": "sha256:2222222222",
        "platform": {"architecture": "armhfp", "os": "linux"},
    },
    {
        "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
        "size": 429,
        "digest": "sha256:5555555555",
        "platform": {"architecture": "amd64", "os": "linux"},
    },
]


def test_init():
    merger = manifest_list_merger.ManifestListMerger("quay.io/src/image:1", "quay.io/dest/image:1")
    assert merger.src_image == "quay.io/src/image:1"
    assert merger.dest_image == "quay.io/dest/image:1"
    assert merger._src_quay_client is None
    assert merger._dest_quay_client is None


def test_init_create_client():
    merger = manifest_list_merger.ManifestListMerger(
        "quay.io/src/image:1",
        "quay.io/dest/image:1",
        "src.quay.io",
        "src-user",
        "src-pass",
        "dest-user",
        "dest-pass",
        host="stage.quay.io",
    )
    assert merger.src_image == "quay.io/src/image:1"
    assert merger.dest_image == "quay.io/dest/image:1"
    assert isinstance(merger._src_quay_client, quay_client.QuayClient)
    assert isinstance(merger._dest_quay_client, quay_client.QuayClient)
    assert merger._src_quay_client.username == "src-user"
    assert merger._src_quay_client.password == "src-pass"
    assert merger._src_quay_client.session.hostname == "src.quay.io"
    assert merger._dest_quay_client.username == "dest-user"
    assert merger._dest_quay_client.password == "dest-pass"
    assert merger._dest_quay_client.session.hostname == "stage.quay.io"


def test_set_client():
    merger = manifest_list_merger.ManifestListMerger("quay.io/src/image:1", "quay.io/dest/image:1")
    assert merger.src_image == "quay.io/src/image:1"
    assert merger.dest_image == "quay.io/dest/image:1"
    assert merger._src_quay_client is None
    assert merger._dest_quay_client is None

    src_client = quay_client.QuayClient("src-user", "src-pass")
    dest_client = quay_client.QuayClient("dest-user", "dest-pass")
    merger.set_quay_clients(src_client, dest_client)
    assert src_client == merger._src_quay_client
    assert dest_client == merger._dest_quay_client


def test_get_missing_architectures():
    merger = manifest_list_merger.ManifestListMerger("quay.io/src/image:1", "quay.io/dest/image:1")
    missing = merger.get_missing_architectures(new_ml, old_ml)

    assert missing == expected_missing_archs


def test_add_missing_architectures():
    merger = manifest_list_merger.ManifestListMerger("quay.io/src/image:1", "quay.io/dest/image:1")
    merged_manifest_list = merger._add_missing_architectures(new_ml, expected_missing_archs)

    merged_manifest_list["manifests"].sort(key=lambda manifest: manifest["digest"])
    merged_ml["manifests"].sort(key=lambda manifest: manifest["digest"])
    assert merged_manifest_list == merged_ml


def test_merge_manifest_lists_success():
    merger = manifest_list_merger.ManifestListMerger(
        "quay.io/src/image:1",
        "quay.io/dest/image:1",
        "src-quay.io",
        "src-user",
        "src-pass",
        "dest-user",
        "dest-pass",
    )

    with requests_mock.Mocker() as m:
        m.get(
            "https://src-quay.io/v2/src/image/manifests/1",
            json=new_ml,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/dest/image/manifests/1",
            json=old_ml,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.put("https://quay.io/v2/dest/image/manifests/1", status_code=200)

        merger.merge_manifest_lists()
        assert m.call_count == 3
        sent_ml = m.request_history[-1].json()
        sent_ml["manifests"].sort(key=lambda manifest: manifest["digest"])
        expected_ml = deepcopy(merged_ml)
        expected_ml["manifests"].sort(key=lambda manifest: manifest["digest"])

        assert sent_ml == expected_ml


def test_merge_manifest_lists_missing_client():
    merger = manifest_list_merger.ManifestListMerger("quay.io/src/image:1", "quay.io/dest/image:1")

    with pytest.raises(RuntimeError, match="QuayClient instance must be set"):
        merger.merge_manifest_lists()


def test_merge_selected_architectures():
    merger = manifest_list_merger.ManifestListMerger(
        "quay.io/src/image:1",
        "quay.io/dest/image:1",
        "src-quay.io",
        "src-user",
        "src-pass",
        "dest-user",
        "dest-pass",
    )

    with requests_mock.Mocker() as m:
        m.get(
            "https://src-quay.io/v2/src/image/manifests/1",
            json=new_ml,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/dest/image/manifests/1",
            json=old_ml,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )

        created_ml = merger.merge_manifest_lists_selected_architectures(["arm64", "s390x"])
        assert m.call_count == 2
        created_ml["manifests"].sort(key=lambda manifest: manifest["digest"])
        expected_ml = deepcopy(merged_ml2)
        expected_ml["manifests"].sort(key=lambda manifest: manifest["digest"])

        assert created_ml == expected_ml


@mock.patch("pubtools._quay.quay_client.QuayClient._authenticate_quay")
def test_merge_selected_architectures_no_dest_manifest(mock_authenticate):
    merger = manifest_list_merger.ManifestListMerger(
        "quay.io/src/image:1",
        "quay.io/dest/image:1",
        "src-quay.io",
        "src-user",
        "src-pass",
        "dest-user",
        "dest-pass",
    )

    with requests_mock.Mocker() as m:
        m.get(
            "https://src-quay.io/v2/src/image/manifests/1",
            json=new_ml,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/dest/image/manifests/1",
            [
                {"headers": {"some-header": "value"}, "status_code": 401},
                {"status_code": 401},
            ],
        )

        created_ml = merger.merge_manifest_lists_selected_architectures(["arm64", "s390x"])
        assert m.call_count == 3
        created_ml["manifests"].sort(key=lambda manifest: manifest["digest"])
        expected_ml = deepcopy(merged_ml3)
        expected_ml["manifests"].sort(key=lambda manifest: manifest["digest"])

        assert created_ml == expected_ml


def test_merge_selected_architectures_raises_unrelated_error():
    merger = manifest_list_merger.ManifestListMerger(
        "quay.io/src/image:1",
        "quay.io/dest/image:1",
        "src-quay.io",
        "src-user",
        "src-pass",
        "dest-user",
        "dest-pass",
    )

    with requests_mock.Mocker() as m:
        m.get(
            "https://src-quay.io/v2/src/image/manifests/1",
            json=new_ml,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/dest/image/manifests/1",
            status_code=500,
        )

        with pytest.raises(requests.exceptions.HTTPError, match=".*500 Server Error.*"):
            merger.merge_manifest_lists_selected_architectures(["arm64", "s390x"])
        assert m.call_count == 2


def test_merge_manifest_lists_selected_architectures_missing_client():
    merger = manifest_list_merger.ManifestListMerger("quay.io/src/image:1", "quay.io/dest/image:1")

    with pytest.raises(RuntimeError, match="QuayClient instance must be set"):
        merger.merge_manifest_lists_selected_architectures(["arm64", "s390x"])



import mock
import pytest

from pubtools._quay.signer_wrapper import SignerWrapper, SigningError, SignEntry


def test_signer_wrapper_entry_point():
    with mock.patch("pkg_resources.load_entry_point") as mock_load_entry_point:
        sw = SignerWrapper(config_file="fake-config-file", settings={})
        sw.entry_point()
        mock_load_entry_point.assert_called_once_with("signer", "group", "signer")


def test_signer_remove_signatures():
    with mock.patch(
        "pubtools._quay.signer_wrapper.SignerWrapper._remove_signatures"
    ) as mock_remove:
        sw = SignerWrapper(config_file="fake-config-file", settings={})
        sw.remove_signatures(1)
        mock_remove.assert_called_once_with(
            1,
        )


def test_sign_containers_failed():
    with mock.patch("pkg_resources.load_entry_point") as mock_load_entry_point:
        sw = SignerWrapper(config_file="fake-config-file", settings={})
        sw.entry_point()
        mock_load_entry_point.return_value = {
            "signer_result": {"status": "error", "error_message": "fake-error-message"}
        }
        with pytest.raises(SigningError):
            sw._sign_containers(
                [
                    SignEntry(
                        pub_reference="",
                        reference="fake-reference",
                        digest="fake-digest",
                        signing_key="fake-signing-key",
                        arch="amd64",
                        repo="fake-repo",
                    )
                ]
            )



import os
import logging
import argparse
from typing import Any, Optional, List, Dict

from pubtools.pluggy import pm, task_context

from .quay_client import QuayClient
from .quay_api_client import QuayApiClient
from .utils.misc import (
    setup_arg_parser,
    add_args_env_variables,
    get_internal_container_repo_name,
)
from .item_processor import item_processor_for_internal_data, VirtualPushItem
from .signer_wrapper import SIGNER_BY_LABEL

LOG = logging.getLogger("pubtools.quay")

REMOVE_REPO_ARGS = {
    ("--repositories",): {
        "help": "External repositories to remove as CSV.",
        "required": True,
        "type": str,
    },
    ("--quay-host",): {
        "help": "quay host name",
        "required": False,
        "type": str,
        "default": "quay.io",
    },
    ("--quay-org",): {
        "help": "Quay organization in which repositories reside.",
        "required": True,
        "type": str,
    },
    ("--quay-api-token",): {
        "help": "OAuth token for Quay REST API.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_API_TOKEN",
    },
    ("--quay-user",): {
        "help": "Username for Quay login.",
        "required": True,
        "type": str,
    },
    ("--quay-password",): {
        "help": "Password for Quay. Can be specified by env variable QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_PASSWORD",
    },
    ("--pyxis-server",): {
        "help": "Pyxis service hostname",
        "required": True,
        "type": str,
    },
    ("--pyxis-ssl-crtfile",): {
        "help": "Path to .crt file for the SSL authentication",
        "required": True,
        "type": str,
    },
    ("--pyxis-ssl-keyfile",): {
        "help": "Path to .key file for the SSL authentication",
        "required": True,
        "type": str,
    },
    ("--pyxis-request-threads",): {
        "help": "Maximum number of threads to use for parallel pyxis request",
        "required": False,
        "default": 7,
        "type": int,
    },
    ("--signers",): {
        "help": "Comma separated list of signerrs",
        "required": False,
        "type": str,
        "default": "",
    },
    ("--signer-configs",): {
        "help": "Comma separated list of signerrs",
        "required": False,
        "type": str,
        "default": "",
    },
}


def remove_repositories(repositories: str, settings: Dict[str, Any]) -> None:
    """
    Remove Quay repository.

    Args:
        repositories (str):
            External repositories to remove. Comma separated values.
        settings (dict):
            Settings dictionary with following keys:
                quay_org (str):
                    Quay organization in which repositories reside.
                quay_api_token (str):
                    OAuth token for authentication of Quay REST API.
                quay_user (str):
                    Quay username for Docker HTTP API.
                quay_password (str):
                    Quay password for Docker HTTP API.
                pyxis_server (str):
                    Pyxis service hostname:
                pyxis_ssl_crtfile (str):
                    Path to .crt file for SSL authentication.
                pyxis_ssl_keyfile (str):
                    Path to .key file for SSL authentication.
                pyxis_request_threads:
                    Maximum number of threads to use for parallel pyxis request.
    """
    parsed_repositories = repositories.split(",")

    LOG.info("Removing repositories '{0}'".format(repositories))
    quay_api_client = QuayApiClient(settings["quay_api_token"])
    quay_client = QuayClient(settings["quay_user"], settings["quay_password"], "quay.io")
    item_processor = item_processor_for_internal_data(
        quay_client, "quay.io", [""], 5, settings["quay_org"]
    )
    item_processor.extractor.full_extract = True
    # Remove repository doesn't work with push item by default, therefore we create VirtualPushItem
    # to support existing code to generate needed repository data.
    item = VirtualPushItem(
        metadata={"tags": {repo: [] for repo in parsed_repositories}},
        repos={repo: [] for repo in parsed_repositories},
    )
    existing_manifests = item_processor.generate_all_existing_manifests_metadata(item)
    signers = settings["signers"].split(",")
    signer_configs = settings["signer_configs"].split(",")
    outdated_manifests = []
    for repo, tag, mad in existing_manifests:
        if not mad:
            continue
        outdated_manifests.append((mad.digest, tag, repo))

    signer_settings = {k: v for k, v in settings.items() if k not in ["quay_org"]}
    signer_settings["quay_namespace"] = settings["quay_org"]
    signer_settings["dest_quay_api_token"] = os.environ.get("QUAY_API_TOKEN")
    signer_settings["quay_host"] = "quay.io"

    for n, signer in enumerate(signers):
        signercls = SIGNER_BY_LABEL[signer]
        if not signercls.pre_push:
            continue
        signer = signercls(config_file=signer_configs[0], settings=signer_settings)
        signer.remove_signatures(outdated_manifests, _exclude=[])

    for repository in parsed_repositories:
        internal_repo = "{0}/{1}".format(
            settings["quay_org"], get_internal_container_repo_name(repository)
        )
        quay_api_client.delete_repository(internal_repo)

    LOG.info("Repositories have been removed")

    pm.hook.quay_repositories_removed(repository_ids=sorted(parsed_repositories))


def setup_args() -> argparse.ArgumentParser:
    """Set up argparser without extra parameters, this method is used for auto doc generation."""
    return setup_arg_parser(REMOVE_REPO_ARGS)


def remove_repositories_main(sysargs: Optional[List[str]] = None) -> None:
    """Entrypoint for removing repositories."""
    logging.basicConfig(level=logging.INFO)

    parser = setup_args()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, REMOVE_REPO_ARGS)

    if not args.quay_api_token:
        raise ValueError("--quay-api-token must be specified")
    if not args.quay_password:
        raise ValueError("--quay-password must be specified")

    kwargs = args.__dict__

    with task_context():
        repositories = kwargs.pop("repositories")
        remove_repositories(repositories, kwargs)



import json
import logging
import mock
import pytest
import requests_mock

from pubtools._quay import quay_client
from pubtools._quay import image_untagger
from .utils.misc import sort_dictionary_sortable_values, compare_logs

# flake8: noqa: E501

try:
    UNUSED = type("Unused", ("object",), {})
except TypeError:
    UNUSED = type("Unused", (), {})


def setup_untagger(
    references,
    token=UNUSED,
    quay_user=UNUSED,
    quay_password=UNUSED,
    host=UNUSED,
    remove_last=UNUSED,
):
    _token = "some-token"
    _quay_user = "some-user"
    _quay_password = "some-password"
    _host = "stage.quay.io/"
    untagger = image_untagger.ImageUntagger(
        references,
        token if token is not UNUSED else _token,
        remove_last if remove_last is not UNUSED else False,
        quay_user if _quay_user is not UNUSED else _quay_user,
        quay_password if quay_password is not UNUSED else _quay_password,
        host if host is not UNUSED else _host,
    )
    return untagger


def register_manifest_url(mocker, repo, manifest, data, mlist=False):
    mocker.get(
        "https://stage.quay.io/v2/name/%s/manifests/%s" % (repo, manifest),
        text=json.dumps(data, sort_keys=True),
        headers=(
            {"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"}
            if mlist
            else {"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"}
        ),
    )


def register_tags_api(mocker, repo, data):
    mocker.get(
        "https://stage.quay.io/v2/name/%s/tags/list" % repo,
        json=data,
    )


@mock.patch("pubtools._quay.image_untagger.QuayClient")
@mock.patch("pubtools._quay.image_untagger.QuayApiClient")
def test_init_success(mock_quay_api_client, mock_quay_client):
    references = ["stage.quay.io/name/repo1:1", "stage.quay.io/name/repo2:2"]
    token = "some-token"
    quay_user = "some-user"
    quay_password = "some-password"
    host = "stage.quay.io/"
    untagger = image_untagger.ImageUntagger(
        references, token, False, quay_user, quay_password, host
    )

    assert untagger.references == references
    assert untagger.host == host[:-1]
    assert not untagger.remove_last
    mock_quay_client.assert_called_once_with(quay_user, quay_password, host[:-1])
    mock_quay_api_client.assert_called_once_with(token, host[:-1])


@mock.patch("pubtools._quay.image_untagger.QuayClient")
@mock.patch("pubtools._quay.image_untagger.QuayApiClient")
def test_init_bad_reference(mock_quay_api_client, mock_quay_client):
    references = [
        "stage.quay.io/name/repo1:1",
        "stage.quay.io/name/repo2@sha256:dgdgdg",
    ]
    with pytest.raises(ValueError, match=".*must be specified via tag, not digest.*"):
        untagger = setup_untagger(references)


@mock.patch("pubtools._quay.image_untagger.QuayClient")
@mock.patch("pubtools._quay.image_untagger.QuayApiClient")
def test_init_no_docker_client(mock_quay_api_client, mock_quay_client):
    references = ["stage.quay.io/name/repo1:1", "stage.quay.io/name/repo2:2"]
    host = "stage.quay.io/"
    token = "some-token"
    untagger = setup_untagger(references, quay_password=None, token=token, host=host)

    mock_quay_client.assert_not_called()
    mock_quay_api_client.assert_called_once_with(token, host[:-1])
    assert untagger._quay_client is None

    client = quay_client.QuayClient("user", "pass")
    untagger.set_quay_client(client)
    assert client == untagger._quay_client


@mock.patch("pubtools._quay.image_untagger.QuayClient")
@mock.patch("pubtools._quay.image_untagger.QuayApiClient")
def test_repo_tag_mapping(mock_quay_api_client, mock_quay_client):
    references = [
        "stage.quay.io/name/repo1:1",
        "stage.quay.io/name/repo1:2",
        "stage.quay.io/name/repo2:2",
        "stage.quay.io/name2/repo1:1",
        "stage.quay.io/name2/repo1:3",
    ]
    untagger = setup_untagger(references)

    repo_tag_mapping = untagger.get_repository_tags_mapping()
    assert repo_tag_mapping == {
        "name/repo1": ["1", "2"],
        "name/repo2": ["2"],
        "name2/repo1": ["1", "3"],
    }


def test_tag_digest_mappings(
    manifest_list_data,
    v2s2_manifest_data,
    common_tag_digest_mapping,
    common_digest_tag_mapping,
):
    references = [
        "stage.quay.io/name/repo1:1",
        "stage.quay.io/name/repo1:2",
        "stage.quay.io/name/repo1:3",
        "stage.quay.io/name/repo1:4",
    ]
    repo_tags = {"name": "repo1", "tags": ["1", "2", "3", "4"]}
    untagger = setup_untagger(references)
    with requests_mock.Mocker() as m:
        register_tags_api(m, "repo1", repo_tags)
        register_manifest_url(m, "repo1", "1", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "2", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "3", v2s2_manifest_data)
        register_manifest_url(m, "repo1", "4", v2s2_manifest_data)
        tag_digest_mapping, digest_tag_mapping = untagger.construct_tag_digest_mappings(
            "name/repo1"
        )
        expected_tag_digest_mapping = common_tag_digest_mapping
        expected_digest_tag_mapping = common_digest_tag_mapping

        sort_dictionary_sortable_values(tag_digest_mapping)
        sort_dictionary_sortable_values(expected_tag_digest_mapping)
        sort_dictionary_sortable_values(digest_tag_mapping)
        sort_dictionary_sortable_values(expected_digest_tag_mapping)

        assert tag_digest_mapping == expected_tag_digest_mapping
        assert digest_tag_mapping == expected_digest_tag_mapping
        assert m.call_count == 13


@mock.patch("pubtools._quay.image_untagger.QuayClient")
@mock.patch("pubtools._quay.image_untagger.QuayApiClient")
def test_get_lost_digests_none(
    mock_quay_api_client,
    mock_quay_client,
    common_tag_digest_mapping,
    common_digest_tag_mapping,
):
    references = [
        "stage.quay.io/name/repo1:1",
    ]
    untagger = setup_untagger(references)

    lost_digests = untagger.get_lost_digests(
        ["1"], common_tag_digest_mapping, common_digest_tag_mapping
    )
    assert lost_digests == []


@mock.patch("pubtools._quay.image_untagger.QuayClient")
@mock.patch("pubtools._quay.image_untagger.QuayApiClient")
def test_get_lost_digests_some(
    mock_quay_api_client,
    mock_quay_client,
    common_tag_digest_mapping,
    common_digest_tag_mapping,
):
    references = [
        "stage.quay.io/name/repo1:1",
        "stage.quay.io/name/repo1:2",
    ]
    untagger = setup_untagger(references)

    lost_digests = untagger.get_lost_digests(
        ["1", "2"], common_tag_digest_mapping, common_digest_tag_mapping
    )
    assert lost_digests == [
        "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
        "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
        "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
        "sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f",
        "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
        "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
    ]


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_get_repo_cosign_images(mock_triangulate):
    references = [
        "stage.quay.io/name/repo1:1",
    ]
    repo_tags = ["1", "2", "sha256-abcd.att", "sha256-abcd.sig"]
    mock_triangulate.side_effect = [
        "stage.quay.io/name/repo1:sha256-abcd.att",
        "stage.quay.io/name/repo1:sha256-abcd.sbom",
        "stage.quay.io/name/repo1:sha256-abcd.sig",
    ]
    untagger = setup_untagger(references)
    cosign_images = untagger.get_repo_cosign_images(
        ["stage.quay.io/name/repo1@sha256:1234abc"], repo_tags
    )
    assert cosign_images == {
        "stage.quay.io/name/repo1:sha256-abcd.att",
        "stage.quay.io/name/repo1:sha256-abcd.sig",
    }

    assert mock_triangulate.call_count == 3


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_get_repo_cosign_images_empty(mock_triangulate):
    references = [
        "stage.quay.io/name/repo1:1",
    ]
    repo_tags = ["1", "2", "sha256-abcd.att", "sha256-abcd.sig"]
    untagger = setup_untagger(references)
    cosign_images = untagger.get_repo_cosign_images([], repo_tags)
    assert cosign_images == set()

    mock_triangulate.assert_not_called()


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_get_repo_cosign_images_tag_images(mock_triangulate):
    references = [
        "stage.quay.io/name/repo1:1",
    ]
    repo_tags = ["1", "2", "sha256-abcd.att", "sha256-abcd.sig"]
    untagger = setup_untagger(references)
    with pytest.raises(ValueError, match=".* are not specified by digest"):
        untagger.get_repo_cosign_images(
            ["stage.quay.io/name/repo1@sha256:1234abc", "stage.quay.io/name/repo1:1"], repo_tags
        )

    mock_triangulate.assert_not_called()


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_get_repo_cosign_images_different_repos(mock_triangulate):
    references = [
        "stage.quay.io/name/repo1:1",
    ]
    repo_tags = ["1", "2", "sha256-abcd.att", "sha256-abcd.sig"]
    untagger = setup_untagger(references)
    with pytest.raises(ValueError, match="Specified images belong to multiple repos"):
        untagger.get_repo_cosign_images(
            ["stage.quay.io/name/repo1@sha256:1234abc", "stage.quay.io/name/repo2@sha256:def5432"],
            repo_tags,
        )

    mock_triangulate.assert_not_called()


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_get_repo_cosign_images_specified_types(mock_triangulate):
    references = [
        "stage.quay.io/name/repo1:1",
    ]
    repo_tags = ["1", "2", "sha256-abcd.att", "sha256-abcd.sig"]
    mock_triangulate.side_effect = [
        "stage.quay.io/name/repo1:sha256-abcd.att",
        "stage.quay.io/name/repo1:sha256-abcd.sbom",
        "stage.quay.io/name/repo1:sha256-abcd.sig",
    ]
    untagger = setup_untagger(references)
    cosign_images = untagger.get_repo_cosign_images(
        ["stage.quay.io/name/repo1@sha256:1234abc"], repo_tags, ["attestation"]
    )
    assert cosign_images == {
        "stage.quay.io/name/repo1:sha256-abcd.att",
    }

    assert mock_triangulate.call_count == 1


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_get_repo_cosign_images_wrong_specified_types(mock_triangulate):
    references = [
        "stage.quay.io/name/repo1:1",
    ]
    repo_tags = ["1", "2", "sha256-abcd.att", "sha256-abcd.sig"]
    mock_triangulate.side_effect = [
        "stage.quay.io/name/repo1:sha256-abcd.att",
        "stage.quay.io/name/repo1:sha256-abcd.sbom",
        "stage.quay.io/name/repo1:sha256-abcd.sig",
    ]
    untagger = setup_untagger(references)
    with pytest.raises(ValueError, match="Unknown cosign image types.*"):
        untagger.get_repo_cosign_images(
            ["stage.quay.io/name/repo1@sha256:1234abc"], repo_tags, ["artestation", "sbomb"]
        )

    assert mock_triangulate.call_count == 0


@mock.patch("pubtools._quay.image_untagger.ImageUntagger.get_repo_cosign_images")
def test_untag_images_no_lost_digests(
    mock_get_repo_cosign_images, manifest_list_data, v2s2_manifest_data, caplog
):
    caplog.set_level(logging.INFO)
    references = [
        "stage.quay.io/name/repo1:1",
    ]
    repo_tags = {"name": "repo1", "tags": ["1", "2", "3", "4"]}
    mock_get_repo_cosign_images.return_value = set()
    untagger = setup_untagger(references)
    with requests_mock.Mocker() as m:
        register_tags_api(m, "repo1", repo_tags)
        register_manifest_url(m, "repo1", "1", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "2", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "3", v2s2_manifest_data)
        register_manifest_url(m, "repo1", "4", v2s2_manifest_data)
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/1")
        lost_images = untagger.untag_images()

        assert lost_images == []
        assert m.call_count == 14

        expected_logs = [
            "Gathering tags and digests of repository 'name/repo1'",
            "No images will be lost by this untagging operation",
            "Removing tag '1' from repository 'name/repo1'",
        ]
        compare_logs(caplog, expected_logs)


@mock.patch("pubtools._quay.image_untagger.ImageUntagger.get_repo_cosign_images")
def test_untag_images_lost_digests_error(
    mock_get_repo_cosign_images, manifest_list_data, v2s2_manifest_data, caplog
):
    caplog.set_level(logging.INFO)
    references = [
        "stage.quay.io/name/repo1:1",
        "stage.quay.io/name/repo1:2",
    ]
    repo_tags = {"name": "repo1", "tags": ["1", "2", "3", "4"]}
    mock_get_repo_cosign_images.return_value = set()
    untagger = setup_untagger(references)
    with requests_mock.Mocker() as m:
        register_tags_api(m, "repo1", repo_tags)
        register_manifest_url(m, "repo1", "1", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "2", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "3", v2s2_manifest_data)
        register_manifest_url(m, "repo1", "4", v2s2_manifest_data)
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/1")

        expected_err_msg = (
            "Following images .*"
            ".*stage.quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f.*"
            ".*stage.quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee.*"
            ".*stage.quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9.*"
            ".*stage.quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c.*"
            ".*stage.quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb.*"
            ".*stage.quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd.*"
        )

        with pytest.raises(ValueError, match=expected_err_msg):
            untagger.untag_images()


@mock.patch("pubtools._quay.image_untagger.ImageUntagger.get_repo_cosign_images")
def test_untag_images_lost_digests_remove_anyway(
    mock_get_repo_cosign_images, manifest_list_data, v2s2_manifest_data, caplog
):
    caplog.set_level(logging.INFO)
    references = [
        "stage.quay.io/name/repo1:1",
        "stage.quay.io/name/repo1:2",
    ]
    repo_tags = {"name": "repo1", "tags": ["1", "2", "3", "4"]}
    untagger = setup_untagger(references, remove_last=True)
    mock_get_repo_cosign_images.return_value = set()
    with requests_mock.Mocker() as m:
        register_tags_api(m, "repo1", repo_tags)
        register_manifest_url(m, "repo1", "1", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "2", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "3", v2s2_manifest_data)
        register_manifest_url(m, "repo1", "4", v2s2_manifest_data)
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/1")
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/2")
        lost_images = untagger.untag_images()

        expected_lost_images = [
            "stage.quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f",
            "stage.quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "stage.quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "stage.quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "stage.quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "stage.quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
        ]

        assert sorted(lost_images) == sorted(expected_lost_images)
        assert m.call_count == 16

        assert mock_get_repo_cosign_images.call_count == 2
        assert mock_get_repo_cosign_images.call_args_list[0] == mock.call(
            expected_lost_images, repo_tags["tags"]
        )
        assert mock_get_repo_cosign_images.call_args_list[1] == mock.call(
            [], repo_tags["tags"], ["signature"]
        )

        expected_logs = [
            "Gathering tags and digests of repository 'name/repo1'",
            "Following images won't be referencable by tag: "
            ".*stage.quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f.*"
            ".*stage.quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee.*"
            ".*stage.quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9.*"
            ".*stage.quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c.*"
            ".*stage.quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb.*"
            ".*stage.quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd.*",
            "Removing tag '1' from repository 'name/repo1'",
            "Removing tag '2' from repository 'name/repo1'",
        ]
        compare_logs(caplog, expected_logs)


@mock.patch("pubtools._quay.image_untagger.ImageUntagger.get_repo_cosign_images")
def test_untag_images_lost_digests_cosign_images(
    mock_get_repo_cosign_images, manifest_list_data, v2s2_manifest_data, caplog
):
    caplog.set_level(logging.INFO)
    references = [
        "stage.quay.io/name/repo1:1",
        "stage.quay.io/name/repo1:2",
    ]
    repo_tags = {"name": "repo1", "tags": ["1", "2", "3", "4"]}
    untagger = setup_untagger(references, remove_last=True)
    mock_get_repo_cosign_images.side_effect = [
        {
            "stage.quay.io/name/repo1:sha256-abcdef.att",
            "stage.quay.io/name/repo1:sha256-abcdef.sig",
        },
        {
            "stage.quay.io/name/repo1:sha256-fedcba.sig",
        },
    ]
    with requests_mock.Mocker() as m:
        register_tags_api(m, "repo1", repo_tags)
        register_manifest_url(m, "repo1", "1", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "2", manifest_list_data, mlist=True)
        register_manifest_url(m, "repo1", "3", v2s2_manifest_data)
        register_manifest_url(m, "repo1", "4", v2s2_manifest_data)
        register_manifest_url(m, "repo1", "sha256-abcdef.att", v2s2_manifest_data)
        register_manifest_url(m, "repo1", "sha256-abcdef.sig", manifest_list_data)
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/1")
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/2")
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/sha256-abcdef.att")
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/sha256-abcdef.sig")
        m.delete("https://stage.quay.io/api/v1/repository/name/repo1/tag/sha256-fedcba.sig")
        lost_images = untagger.untag_images()

        expected_lost_images = [
            "stage.quay.io/name/repo1:sha256-abcdef.att",
            "stage.quay.io/name/repo1:sha256-abcdef.sig",
            "stage.quay.io/name/repo1:sha256-fedcba.sig",
            "stage.quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f",
            "stage.quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "stage.quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "stage.quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "stage.quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "stage.quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
        ]

        assert sorted(lost_images) == sorted(expected_lost_images)
        assert m.call_count == 23

        assert mock_get_repo_cosign_images.call_count == 2
        assert mock_get_repo_cosign_images.call_args_list[0] == mock.call(
            expected_lost_images[3:], repo_tags["tags"]
        )
        assert mock_get_repo_cosign_images.call_args_list[1] == mock.call(
            [
                "stage.quay.io/name/repo1@sha256:78060d7b3da37ef95fe133f82c0efb0a0c730da9a4178b5767213a1e7a59fff1",
                "stage.quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f",
            ],
            repo_tags["tags"],
            ["signature"],
        )

        expected_logs = [
            "Gathering tags and digests of repository 'name/repo1'",
            "Following images won't be referencable by tag: "
            ".*stage.quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f.*"
            ".*stage.quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee.*"
            ".*stage.quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9.*"
            ".*stage.quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c.*"
            ".*stage.quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb.*"
            ".*stage.quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd.*",
            "Following cosign images won't be referecable by tag: "
            ".*stage.quay.io/name/repo1:sha256-abcdef.att.*"
            ".*stage.quay.io/name/repo1:sha256-abcdef.sig.*"
            ".*stage.quay.io/name/repo1:sha256-fedcba.sig.*",
            "Removing tag '1' from repository 'name/repo1'",
            "Removing tag '2' from repository 'name/repo1'",
            "Removing tag 'sha256-abcdef.att' from repository 'name/repo1'",
            "Removing tag 'sha256-abcdef.sig' from repository 'name/repo1'",
            "Removing tag 'sha256-fedcba.sig' from repository 'name/repo1'",
        ]
        compare_logs(caplog, expected_logs)


def test_untag_images_missing_client(manifest_list_data, caplog):
    caplog.set_level(logging.INFO)
    references = [
        "stage.quay.io/name/repo1:1",
        "stage.quay.io/name/repo1:2",
    ]
    token = "some-token"
    quay_user = "some-user"
    host = "stage.quay.io/"
    untagger = image_untagger.ImageUntagger(references, token, True, quay_user, host=host)
    with pytest.raises(RuntimeError, match="QuayClient instance must be set"):
        untagger.untag_images()



import abc
from contextlib import contextmanager, redirect_stdout
import logging
import pkg_resources
import tempfile
import json
import io

from typing import Optional, List, Dict, Any, Tuple, Generator, Type

from marshmallow import Schema, fields, EXCLUDE

from .utils.misc import (
    run_entrypoint,
    get_pyxis_ssl_paths,
    # run_in_parallel,
    log_step,
    # FData,
)
from .item_processor import SignEntry

LOG = logging.getLogger("pubtools.quay")


class SigningError(Exception):
    """Error raised when signing fails."""


class NoSchema(Schema):
    """Schema that does not validate anything."""


class MsgSignerSettingsSchema(Schema):
    """Validation schema for messaging signer settings."""

    pyxis_server = fields.String(required=True)
    pyxis_ssl_crtfile = fields.String(required=False)
    pyxis_ssl_keyfile = fields.String(required=False)
    num_thread_pyxis = fields.Integer(required=False, default=7)


class SignerWrapper:
    """Wrapper providing functionality to sign containers with a generic signer."""

    label = "unused"
    pre_push = False

    SCHEMA: Type[Schema] = NoSchema
    entry_point_conf = ["signer", "group", "signer"]

    def __init__(
        self, config_file: Optional[str] = None, settings: Optional[Dict[str, Any]] = None
    ) -> None:
        """Initialize SignerWrapper.

        Args:
            config_file (str): Path to config file for the signer.
            settings (dict): Settings for the signer.
        """
        self.config_file = config_file
        self.settings = settings or {}
        self._ep = None
        self.validate_settings()

    @property
    def entry_point(self) -> Any:
        """Load and return entry point for pubtools-sign project."""
        if self._ep is None:
            self._ep = pkg_resources.load_entry_point(*self.entry_point_conf)
        return self._ep

    def remove_signatures(
        self,
        signatures: List[Tuple[str, str, str]],
        _exclude: Optional[List[Tuple[str, str, str]]] = None,
    ) -> None:
        """Remove signatures from a sigstore."""
        LOG.debug("Removing signatures %s", signatures)
        self._remove_signatures(signatures)

    @abc.abstractmethod
    def _run_remove_signatures(self, signatures_to_remove: List[Any]) -> None:
        pass  # pragma: no cover

    def _remove_signatures(self, signatures_to_remove: List[Any]) -> None:
        """Remove signatures from sigstore.

        This is helper to make testing easier.
        Args:
            signatures_to_remove (list): Signatures to remove.
        """
        self._run_remove_signatures(signatures_to_remove)

    @abc.abstractmethod
    def _run_store_signed(self, signatures: Dict[str, Any]) -> None:
        pass  # pragma: no cover

    def _store_signed(self, signatures: Dict[str, Any]) -> None:
        """Store signatures in sigstore.

        This is helper to make testing easier.
        Args:
            signatures (dict): Signatures to store.
        """
        LOG.debug("Storing signatures %s", signatures)
        self._run_store_signed(signatures)

    def sign_container_opt_args(
        self, sign_entry: List[SignEntry], task_id: Optional[str] = None
    ) -> Dict[str, Any]:
        """Return optional arguments for signing a container.

        Args:
            sign_entries (List[SignEntry]): List of SignEntry.
            task_id (str): Task ID to identify the signing task if needed.

        Returns:
            dict: Optional arguments for signing a container.
        """
        return {}

    def _sign_containers(
        self,
        sign_entries: List[SignEntry],
        task_id: Optional[str] = None,
    ) -> None:
        """Sign a specific chunk of references and digests with given signing key.

        Args:
            sign_entries (List[SignEntry]): Chunk of SignEntry to sign.
            task_id (str): Task ID to identify the signing task if needed.
        """
        for sign_entry in sign_entries:
            LOG.debug(
                "Signing container %s %s %s",
                sign_entry.reference,
                sign_entry.digest,
                sign_entry.signing_key,
            )
        if not sign_entries:
            return
        sign_entry = sign_entries[0]
        opt_args = self.sign_container_opt_args(sign_entries, task_id)
        signed = self.entry_point(
            config_file=self.config_file,
            signing_key=sign_entry.signing_key,
            reference=[x.reference for x in sign_entries if x],
            digest=[x.digest for x in sign_entries if x],
            **opt_args,
        )
        if signed["signer_result"]["status"] != "ok":
            raise SigningError(signed["signer_result"]["error_message"])
        for sign_entry in sign_entries:
            LOG.info(
                "Signed %s(%s) with %s in %s",
                sign_entry.reference,
                sign_entry.digest,
                sign_entry.signing_key,
                self.label,
            )
        self._store_signed(signed)

    def _filter_to_sign(self, to_sign_entries: List[SignEntry]) -> List[SignEntry]:
        """Filter entries to sign.

        Args:
            to_sign_entries (List[SignEntry]): list of entries to sign.

        Returns:
            List[SignEntry]: list of entries to sign.
        """
        return to_sign_entries

    @log_step("Sign container images")
    def sign_containers(
        self,
        to_sign_entries: List[SignEntry],
        task_id: Optional[str] = None,
    ) -> None:
        """Sign signing entries.

        Entries are sent to signer in chunks of chunk_size size.

        Args:
            to_sign_entries (List[SignEntry]): list of entries to sign.
            task_id (str): optional identifier used in signing process.
            parallelism (int): determines how many entries should be signed in parallel.
        """
        to_sign_entries = self._filter_to_sign(to_sign_entries)
        to_sign_entries_filtered = []
        for sign_entry in to_sign_entries:
            if sign_entry not in to_sign_entries_filtered:
                to_sign_entries_filtered.append(sign_entry)
        with redirect_stdout(io.StringIO()):
            self._sign_containers(to_sign_entries_filtered, task_id)

    def validate_settings(self, settings: Optional[Dict[str, Any]] = None) -> None:
        """Validate provided settings for the SignerWrapper."""
        settings = settings or self.settings
        schema = self.SCHEMA(unknown=EXCLUDE)
        schema.load(settings)


class MsgSignerWrapper(SignerWrapper):
    """Wrapper for messaging signer functionality."""

    label = "msg_signer"
    pre_push = True

    entry_point_conf = ["pubtools-sign", "modules", "pubtools-sign-msg-container-sign"]

    MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST = 50
    SCHEMA = MsgSignerSettingsSchema

    def _filter_to_sign(self, to_sign_entries: List[SignEntry]) -> List[SignEntry]:
        to_sign_digests = [x.digest for x in to_sign_entries]
        existing_signatures = [esig for esig in self._fetch_signatures(to_sign_digests)]
        existing_signatures_drk = {
            (x["manifest_digest"], x["reference"], x["sig_key_id"]) for x in existing_signatures
        }
        ret = []
        for tse in to_sign_entries:
            if (tse.digest, tse.reference, tse.signing_key) not in existing_signatures_drk:
                ret.append(tse)
        return ret

    def sign_container_opt_args(
        self, sign_entries: List[SignEntry], task_id: Optional[str] = None
    ) -> Dict[str, Any]:
        """Return optional arguments for signing a container.

        Args:
            sign_entries (List[SignEntry]): List of SignEntry.
            task_id (str): Task ID to identify the signing task if needed.

        Returns:
            dict: Optional arguments for signing a container.
        """
        return {k: v for k, v in [("task_id", task_id)] if v is not None}

    @contextmanager
    def _save_signatures_file(self, signatures: List[Dict[str, Any]]) -> Generator[Any, None, None]:
        """Save signatures to a temporary file and yield the file."""
        with tempfile.NamedTemporaryFile(
            mode="w", prefix="pubtools_quay_upload_signatures_"
        ) as signature_file:
            json.dump(signatures, signature_file)
            signature_file.flush()
            yield signature_file

    def _fetch_signatures(
        self, manifest_digests: List[str]
    ) -> Generator[Dict[str, Any], None, None]:
        """Fetch signatures from sigstore.

        Args:
            manifest_digests (list): Manifest digests to fetch signatures for.
        Returns:
            List[Dict[str, Any]]: List of fetched signatures.
        """
        cert, key = get_pyxis_ssl_paths(self.settings)
        chunk_size = self.MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST
        manifest_digests = sorted(list(set(manifest_digests)))

        args = ["--pyxis-server", self.settings["pyxis_server"]]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]
        args += ["--request-threads", str(self.settings.get("num_thread_pyxis", 7))]

        for chunk_start in range(0, len(manifest_digests), chunk_size):
            chunk = manifest_digests[chunk_start : chunk_start + chunk_size]  # noqa: E203

            args = ["--pyxis-server", self.settings["pyxis_server"]]
            args += ["--pyxis-ssl-crtfile", cert]
            args += ["--pyxis-ssl-keyfile", key]

            with tempfile.NamedTemporaryFile(
                mode="w", prefix="pubtools_quay_get_signatures_"
            ) as signature_fetch_file:
                if manifest_digests:
                    json.dump(chunk, signature_fetch_file)
                    signature_fetch_file.flush()
                    args += ["--manifest-digest", "@{0}".format(signature_fetch_file.name)]

                env_vars: Dict[Any, Any] = {}
                chunk_results = run_entrypoint(
                    ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-signatures"),
                    "pubtools-pyxis-get-signatures",
                    args,
                    env_vars,
                )

            for result in chunk_results:
                yield result

    def _run_store_signed(self, signed_results: Dict[str, Any]) -> None:
        """
        Upload signatures to Pyxis by using a pubtools-pyxis entrypoint.

        Data required for a Pyxis POST request:
        - manifest_digest
        - reference
        - repository
        - sig_key_id
        - signature_data

        Signatures are uploaded in batches.

        Args:
            signed_results: (Dict[str, Any]):
                Dictionary of {"signer_result":..., "operation_results":..., "signing_key":...}"}
                holding signed manifest claims data
        """
        LOG.info("Sending new signatures to Pyxis")

        signatures: List[Dict[str, Any]] = []
        for reference, op_res in zip(
            signed_results["operation"]["references"], signed_results["operation_results"]
        ):
            signatures.append(
                {
                    "manifest_digest": op_res[0]["msg"]["manifest_digest"],
                    "reference": reference,
                    "repository": op_res[0]["msg"]["repo"],
                    "sig_key_id": signed_results["signing_key"],
                    "signature_data": op_res[0]["msg"]["signed_claim"],
                }
            )

        for sig in signatures:
            LOG.debug(
                f"Uploading new signature. Reference: {sig['reference']}, "
                f"Repository: {sig['repository']}, "
                f"Digest: {sig['manifest_digest']}, "
                f"Key: {sig['sig_key_id']}"
            )

        cert, key = get_pyxis_ssl_paths(self.settings)

        args = ["--pyxis-server", self.settings["pyxis_server"]]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]
        args += ["--request-threads", str(self.settings.get("num_thread_pyxis", 7))]

        with self._save_signatures_file(signatures) as signature_file:
            args += ["--signatures", "@{0}".format(signature_file.name)]
            LOG.info("Uploading {0} new signatures".format(len(signatures)))
            env_vars: Dict[Any, Any] = {}
            run_entrypoint(
                ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-upload-signatures"),
                "pubtools-pyxis-upload-signature",
                args,
                env_vars,
                False,
            )

    def _run_remove_signatures(self, signatures_to_remove: List[str]) -> None:
        """Remove signatures from the sigstore.

        Args:
            signatures_to_remove (List[str]): List of signatures to remove.
        """
        cert, key = get_pyxis_ssl_paths(self.settings)
        args = []
        args = ["--pyxis-server", self.settings["pyxis_server"]]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]
        args += ["--request-threads", str(self.settings.get("num_thread_pyxis", 7))]

        with tempfile.NamedTemporaryFile(mode="w") as temp:
            json.dump(signatures_to_remove, temp)
            temp.flush()

            args += ["--ids", "@%s" % temp.name]

            env_vars: Dict[Any, Any] = {}
            run_entrypoint(
                ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-delete-signatures"),
                "pubtools-pyxis-delete-signatures",
                args,
                env_vars,
            )

    def _filter_to_remove(
        self,
        signatures: List[Tuple[str, str, str]],
        _exclude: Optional[List[Tuple[str, str, str]]] = None,
    ) -> List[str]:
        """Filter signatures to remove.

        Args:
            signatures (List[Tuple[str, str, str]]): List of (digest, tag, repository)
            tuples of signautres to remove.
            _exclude (Optional[List[Tuple[str, str, str]]]): List of  (digest, tag, repository)
            tuples of signautres to keep.
        """
        exclude = _exclude or []
        signatures_to_remove = list(self._fetch_signatures([x[0] for x in signatures]))
        sig_ids_to_remove = []
        for existing_signature in signatures_to_remove:
            if (
                existing_signature["manifest_digest"],
                existing_signature["reference"].split(":")[-1],
                existing_signature["repository"],
            ) in signatures and (
                existing_signature["manifest_digest"],
                existing_signature["reference"],
                existing_signature["repository"],
            ) not in exclude:
                sig_ids_to_remove.append(existing_signature["_id"])
                LOG.debug(
                    f"Removing signature. Reference: {existing_signature['reference']}, "
                    f"Repository: {existing_signature['repository']}, "
                    f"Digest: {existing_signature['manifest_digest']}, "
                    f"Key: {existing_signature['sig_key_id']}"
                )
        return sig_ids_to_remove

    @log_step("Remove outdated signatures")
    def remove_signatures(
        self,
        signatures: List[Tuple[str, str, str]],
        _exclude: Optional[List[Tuple[str, str, str]]] = None,
    ) -> None:
        """Remove signatures from sigstore.

        Args:
            signatures (list): List of tuples containing (digest, reference, repository) of
            signatures to remove.
            exclude (Optional[List[Tuple[str, str, str]]]): List of  (digest, tag, repository)
            tuples of signautres to keep.
        """
        _signatures = list(signatures)
        to_remove = self._filter_to_remove(_signatures, _exclude=_exclude)
        self._remove_signatures(to_remove)


class CosignSignerSettingsSchema(Schema):
    """Validation schema for cosign signer settings."""

    quay_namespace = fields.String(required=True)
    quay_host = fields.String(required=True)
    dest_quay_user = fields.String(required=True)
    dest_quay_password = fields.String(required=True)
    dest_quay_api_token = fields.String(required=True)
    signing_chunk_size = fields.Integer(required=False, default=100)
    signing_parallelism = fields.Integer(required=False, default=10)


class CosignSignerWrapper(SignerWrapper):
    """Wrapper for cosign signer functionality."""

    label = "cosign_signer"
    pre_push = False

    entry_point_conf = ["pubtools-sign", "modules", "pubtools-sign-cosign-container-sign"]

    SCHEMA = CosignSignerSettingsSchema

    # TODO: Uncomment when cosign signature removal is enabled
    # def _list_signatures(self, repository: str, digest: str) -> List[Tuple[str, str]]:
    #     """List cosign signatures for given repository.
    #
    #     This methods runs pubtools-sign-cosign-container-list entrypoint which is expected to
    #     return list of full references to signature tags in format sha256-<digest>.sig
    #
    #     Args:
    #         repository (str): Repository to list signatures for.
    #         tag (str): Tag to list signatures for.
    #     Returns:
    #         List[Tuple[str, str]]: List of (repository, signature tag) tuples
    #         for existing signatures.
    #     """
    #     quay_client = QuayClient(
    #         self.settings["dest_quay_user"],
    #         self.settings["dest_quay_password"],
    #         self.settings["quay_host"],
    #     )
    #     try:
    #         ml = quay_client.get_manifest(
    #             f"{self.settings['quay_host'].rstrip('/')}/"
    #             f"{self.settings['quay_namespace']}/{repository.replace('/','----')}@{digest}",
    #             media_type=QuayClient.MANIFEST_LIST_TYPE,
    #         )
    #     except ManifestTypeError:
    #         ml = None
    #     if ml:
    #         full_references = [
    #             f"{self.settings['quay_host'].rstrip('/')}"
    #             f"/{self.settings['quay_namespace']}/{repository.replace('/','----')}@{digest}"
    #         ]
    #         for manifest in cast(ManifestList, ml)["manifests"]:
    #             full_references.append(
    #                 (
    #                     f"{self.settings['quay_host'].rstrip('/')}/"
    #                     + f"{self.settings['quay_namespace']}/{repository.replace('/','----')}"
    #                     + f"@{manifest['digest']}"
    #                 )
    #             )
    #     else:
    #         full_references = [
    #             (
    #                 f"{self.settings['quay_host'].rstrip('/')}/"
    #                 + f"{self.settings['quay_namespace']}/{repository.replace('/','----')}"
    #                 + f"@{digest}"
    #             )
    #         ]
    #     existing_signatures = []
    #     for full_reference in full_references:
    #         existing_signatures.append(
    #             run_entrypoint(
    #                 ("pubtools-sign", "modules", "pubtools-sign-cosign-signature-list"),
    #                 None,
    #                 [cast(str, self.config_file), full_reference],
    #                 {},
    #             )
    #         )
    #     rets = []
    #     for esig in existing_signatures:
    #         if esig[0] is True:
    #             for sig in esig[1]:
    #                 # if not sig:
    #                 #     continue
    #                 rets.extend(
    #                     [(repository, sig.split(":")[-1].replace(".sig", "").replace("-", ":"))]
    #                 )
    #         else:
    #             LOG.warning("Fetch existing signatures error:" + esig[1])
    #     return rets

    def sign_container_opt_args(
        self, sign_entries: List[SignEntry], task_id: Optional[str] = None
    ) -> Dict[str, Any]:
        """Return optional arguments for signing a container.

        Args:
            sign_entries (List[SignEntry]): List of SignEntry.
            task_id (str): Task ID to identify the signing task if needed.

        Returns:
            dict: Optional arguments for signing a container.
        """
        return {"identity": [sign_entry.pub_reference for sign_entry in sign_entries]}

    def _filter_to_remove(
        self,
        signatures: List[Tuple[str, str, str]],
        _exclude: Optional[List[Tuple[str, str, str]]] = None,
    ) -> List[str]:
        """Filter signatures to remove.

        Args:
            signatures (List[Tuple[str, str, str]]): List of (digest, tag, repository)
            tuples of signautres to remove.
            _exclude (Optional[List[Tuple[str, str, str]]]): List of  (digest, tag, repository)
            tuples of signautres to keep.
        """
        # TODO: Uncomment when cosign signature removal is enabled
        # signatures_to_remove = [(x[2], x[0]) for x in signatures]
        # signatures_to_exclude = [(x[2], x[0]) for x in _exclude or []]
        # existing_signatures = set(
        #     sum(
        #         run_in_parallel(
        #             self._list_signatures,
        #             [FData(args=digest_tag) for digest_tag in signatures_to_remove],
        #         ).values(),
        #         [],
        #     )
        # )
        #
        # to_remove = []
        # for existing_signature in existing_signatures:
        #     if (
        #         existing_signature in signatures_to_remove
        #         and existing_signature not in signatures_to_exclude
        #     ):
        #         to_remove.append(existing_signature)
        #         LOG.debug(
        #             f"Removing signature "
        #             f"Repository: {existing_signature[0]}, "
        #             f"Digest: {existing_signature[1]}, "
        #         )
        # return to_remove
        return []  # pragma: no cover

    def _run_remove_signatures(self, signatures_to_remove: List[Tuple[str, str]]) -> None:
        """Remove signatures from the sigstore.

        Args:
            signatures_to_remove (List[Tuple(str, str)]): List of signatures to remove.
        """
        # TODO: Uncomment when cosign signature removal is enabled
        # qc = QuayApiClient(self.settings["dest_quay_api_token"], host=self.settings["quay_host"])
        # for sig_to_remove in signatures_to_remove:
        #     ref = self.settings["quay_namespace"] + "/" + sig_to_remove[0].replace("/", "----")
        #     sig_tag = sig_to_remove[1].replace(":", "-") + ".sig"
        #     qc.delete_tag(ref, sig_tag)
        pass  # pragma: no cover

    def remove_signatures(
        self,
        signatures: List[Tuple[str, str, str]],
        _exclude: Optional[List[Tuple[str, str, str]]] = None,
    ) -> None:
        """Remove signatures from sigstore.

        Args:
            signatures (list): List of tuples containing (digest, reference, repository) of
            signatures to remove.
            exclude (Optional[List[Tuple[str, str, str]]]): List of  (digest, tag, repository)
            tuples of signautres to keep.
        """
        # XXX: Removal of signatures in cosign is currently disabled. Once cosign will be able to
        # remove specific annotation from a signatures, this method will be updated and enabled.
        return
        # to_remove = self._filter_to_remove(signatures, _exclude=_exclude)
        # self._remove_signatures(to_remove)


SIGNER_BY_LABEL = {
    wrapper.label: wrapper
    for name, wrapper in locals().items()
    if type(wrapper) is type and issubclass(wrapper, SignerWrapper) and wrapper != SignerWrapper
}



from collections import namedtuple
import logging
import sys
import json
from typing import Tuple, List, Dict, Any, cast, Optional

import requests
import urllib3

from .exceptions import BadPushItem, InvalidTargetSettings, InvalidRepository
from .utils.misc import get_internal_container_repo_name, log_step
from .quay_api_client import QuayApiClient
from .quay_client import QuayClient
from .iib_operations import _sign_index_image
from .container_image_pusher import ContainerImagePusher
from .security_manifest_pusher import SecurityManifestPusher

from .operator_pusher import OperatorPusher
from .item_processor import (
    item_processor_for_external_data,
    item_processor_for_internal_data,
    SignEntry,
    ManifestArchDigest,
)
from .utils.misc import parse_index_image
from .signer_wrapper import SIGNER_BY_LABEL

from .utils.misc import (
    get_external_container_repo_name,
    timestamp,
    pyxis_get_repo_metadata,
    set_aws_kms_environment_variables,
    run_in_parallel,
    FData,
)

# TODO: do we want this, or should I remove it?
from urllib3.exceptions import InsecureRequestWarning

urllib3.disable_warnings(InsecureRequestWarning)

LOG = logging.getLogger("pubtools.quay")


class PushDocker:
    """Handle full Docker push workflow."""

    ImageData = namedtuple(
        "ImageData", ["repo", "tag", "v2list_digest", "v2s2_digest", "v2s1_digest"]
    )

    def __init__(
        self,
        push_items: List[Any],
        hub: Any,
        task_id: str,
        target_name: str,
        target_settings: Dict[str, Any],
    ) -> None:
        """
        Initialize.

        Args:
            push_items ([_PushItem]):
                List of push items.
            hub (HubProxy):
                Instance of XMLRPC pub-hub proxy.
            task_id (str):
                task id
            target_name (str):
                Name of the target.
            target_settings (dict):
                Target settings.
        """
        self.push_items = push_items
        self.hub = hub
        self.task_id = task_id
        self.target_name = target_name
        self.target_settings = target_settings

        self.verify_target_settings()

        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")

        self._src_quay_client: Optional[QuayClient] = None
        self._dest_quay_client: Optional[QuayClient] = None
        self._dest_operator_quay_client: Optional[QuayClient] = None
        self._dest_quay_api_client: Optional[QuayApiClient] = None
        self._index_image_quay_client = None
        self.dest_registries = self.target_settings["docker_settings"]["docker_reference_registry"]
        self.dest_registries = (
            self.dest_registries
            if isinstance(self.dest_registries, list)
            else [self.dest_registries]
        )

    @property
    def dest_quay_client(self) -> QuayClient:
        """Create and access QuayClient for dest image."""
        if self._dest_quay_client is None:
            self._dest_quay_client = QuayClient(
                self.target_settings["dest_quay_user"],
                self.target_settings["dest_quay_password"],
                self.quay_host,
            )
        return self._dest_quay_client

    @property
    def src_quay_client(self) -> QuayClient:
        """Create and access QuayClient for source image."""
        if self._src_quay_client is None:
            self._src_quay_client = QuayClient(
                self.target_settings["source_quay_user"],
                self.target_settings["source_quay_password"],
                self.target_settings.get("source_quay_host") or self.quay_host,
            )
        return self._src_quay_client

    @property
    def dest_operator_quay_client(self) -> QuayClient:
        """Create and access QuayClient for dest image."""
        if self._dest_operator_quay_client is None:
            self._dest_operator_quay_client = QuayClient(
                self.target_settings.get(
                    "index_image_quay_user", self.target_settings["dest_quay_user"]
                ),
                self.target_settings.get(
                    "index_image_quay_password", self.target_settings["dest_quay_password"]
                ),
                self.quay_host,
            )
        return self._dest_operator_quay_client

    @property
    def dest_quay_api_client(self) -> QuayApiClient:
        """Create and access QuayApiClient for dest image."""
        if self._dest_quay_api_client is None:
            self._dest_quay_api_client = QuayApiClient(
                self.target_settings["dest_quay_api_token"], self.quay_host
            )
        return self._dest_quay_api_client

    def verify_target_settings(self) -> None:
        """Verify that target settings contains all the necessary data."""
        LOG.info("Verifying the necessary target settings")
        required_settings = [
            "source_quay_user",
            "source_quay_password",
            "dest_quay_user",
            "dest_quay_password",
            "dest_quay_api_token",
            "pyxis_server",
            "quay_namespace",
            "iib_krb_principal",
            "iib_organization",
            "iib_index_image",
            "quay_operator_repository",
            "skopeo_image",
            "docker_settings",
        ]
        for setting in required_settings:
            if setting not in self.target_settings:
                raise InvalidTargetSettings(
                    "'{0}' must be present in the target settings.".format(setting)
                )

        required_docker_settings = ["umb_urls", "docker_reference_registry"]
        for setting in required_docker_settings:
            if setting not in self.target_settings["docker_settings"]:
                raise InvalidTargetSettings(
                    "'{0}' must be present in the docker settings.".format(setting)
                )

    @log_step("Get container push items")
    def get_docker_push_items(self) -> List[Any]:
        """
        Filter push items to only include docker ones.

        Also, Check the validity of these items and raise an exception in case of incorrect data.

        For items having the same pull_url, only one of them is returned to avoid duplication.

        Returns ([ContainerPushItem]):
            Docker push items.
        """
        docker_push_items = []
        url_items: Dict[str, Any] = {}
        for item in self.push_items:
            if item.file_type != "docker":
                LOG.warning("Push item {0} doesn't have 'docker' type, skipping.".format(item))
                continue
            if item.errors:
                raise BadPushItem("Push item {0} contains errors: {1}".format(item, item.errors))
            if not item.metadata.get("pull_data"):
                raise BadPushItem("Push item {0} doesn't contain pull data.".format(item))
            LOG.info("Docker push item found: {0}".format(item))
            pull_url = item.metadata["pull_url"]
            if pull_url in url_items:
                url_items[pull_url].append(item)
            else:
                url_items[pull_url] = [item]
        for _, items in url_items.items():
            non_amd64 = True
            for item in items:
                if item.metadata["arch"] in ["amd64", "x86_64"]:
                    docker_push_items.append(item)
                    non_amd64 = False
                    break
            if non_amd64:
                docker_push_items.append(items[0])

        return docker_push_items

    @log_step("Get operator push items")
    def get_operator_push_items(self) -> List[Any]:
        """
        Filter out push items to only include operator ones.

        Also, Check the validity of these items and raise an exception in case of incorrect data.

        Returns ([ContainerPushItem]):
            Operator push items.
        """
        operator_push_items = []
        for item in self.push_items:
            if item.file_type != "operator":
                LOG.warning("Push item {0} doesn't have 'operator' type, skipping.".format(item))
                continue
            if item.errors:
                raise BadPushItem("Push item {0} contains errors: {1}".format(item, item.errors))
            if not item.metadata.get("op_type"):
                raise BadPushItem("Push item {0} doesn't contain 'op_type'".format(item))
            if item.metadata["op_type"] == "appregistry":
                LOG.warning(
                    "Push item {0} is unsupported legacy (appregistry), skipping".format(item)
                )
                continue
            if item.metadata["op_type"] != "bundle":
                message = "Push item {0} has unknown op_type: '{1}'.".format(
                    item, item.metadata["op_type"]
                )
                raise BadPushItem(message)
            if not item.metadata.get("com.redhat.openshift.versions"):
                msg = "Push item {0} doesn't specify 'com.redhat.openshift.versions'".format(item)
                raise BadPushItem(msg)
            LOG.info("Operator push item found: {0}".format(item))
            operator_push_items.append(item)

        return operator_push_items

    @classmethod
    def check_repos_validity(
        cls, push_items: List[Any], hub: Any, target_settings: Dict[str, Any]
    ) -> None:
        """
        Check if specified repos are valid and pushing to them is allowed.

        Specifically, this method checks if the repo exists in Comet and if it's not deprecated
        If pushing to prod, also check if the repo already exists in stage.

        Args:
            push_items ([ContainerPushItem]):
                Container push items containing the repositories.
            hub (HubProxy):
                Instance of XMLRPC pub-hub proxy.
            target_settings (dict):
                Target settings.
        """
        repos = []
        for item in push_items:
            if item.external_repos:
                repos += item.external_repos.keys()
            else:
                repos += item.repos.keys()

        repos = sorted(list(set(repos)))
        repo_schema = "{namespace}/{repo}"

        # we'll need to get stage namespace from stage target settings
        if "propagated_from" in target_settings:
            stage_target_info = hub.worker.get_target_info(target_settings["propagated_from"])
            stage_namespace = stage_target_info["settings"]["quay_namespace"]
            stage_quay_client = QuayClient(
                stage_target_info["settings"]["dest_quay_user"],
                stage_target_info["settings"]["dest_quay_password"],
            )

        for repo in repos:
            # Only check Pyxis if the option is enabled in target settings
            if target_settings.get("do_repo_deprecation_check", False):
                LOG.info("Checking validity of repository metadata '{0}'".format(repo))
                # Check if repo exists in Pyxis
                try:
                    metadata = pyxis_get_repo_metadata(repo, target_settings)
                    # Check if repo is not deprecated
                    if "Deprecated" in metadata.get("release_categories", []):
                        raise InvalidRepository("Repository {0} is deprecated".format(repo))
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code == 404:
                        LOG.warning(
                            "Metadata of repository '{0}' don't exist in Pyxis".format(repo)
                        )
                    else:
                        raise

            # if we're pushing to prod target, check if repo exists on stage as well
            if "propagated_from" in target_settings:
                internal_repo = get_internal_container_repo_name(repo)
                full_repo = repo_schema.format(namespace=stage_namespace, repo=internal_repo)
                try:
                    stage_quay_client.get_repository_tags(full_repo)
                except requests.exceptions.HTTPError as e:
                    # strangely, if repo doesn't exist, 401 is returned if a robot account is used
                    if e.response.status_code == 404 or e.response.status_code == 401:
                        raise InvalidRepository(
                            "Repository {0} doesn't exist on stage".format(repo)
                        )
                    else:
                        raise

    @log_step("Generate backup mapping")
    def generate_backup_mapping(
        self, push_items: List[Any], all_arches: bool = False
    ) -> Tuple[Dict[ImageData, Tuple[str, str]], List[ImageData]]:
        """
        Create resources which will be used for rollback if something goes wrong during the push.

        Specifically, create two resources: 'backup_tags' and 'rollback_tags'.
        - 'backup_tags' is a mapping of ImageData->manifest, and consists of images which will
        be overwritten. During rollback, tag is made to re-reference the old manifest.
        - 'rollback_tags' is a list of ImageData which don't yet exist. During rollback, they
        will be removed to preserve pre-push state.
        If all_arches is set to true, return all arches for v2s2 and v2s1 manifests

        ImageData is a namedtuple used to assign and access parts of an image in a formatted way.

        Args:
            push_items ([ContainerPushItem]):
                Container push items.
            all_arches: bool
                If set to True include all manifests in results. If False only amd64 are included

        Returns (({ImageData: Tuple[str,str]}, [ImageData])):
            Tuple of backup_tags and rollback_tags
        """
        backup_tags = {}
        rollback_tags = []
        internal_item_processor = item_processor_for_internal_data(
            self.dest_quay_client,
            self.target_settings["quay_host"].rstrip("/"),
            self.dest_registries,
            self.target_settings.get("retry_sleep_time", 5),
            self.target_settings["quay_namespace"],
        )
        external_item_processor = item_processor_for_external_data(
            self.dest_quay_client,
            self.dest_registries,
            self.target_settings.get("retry_sleep_time", 5),
        )
        existing_manifests_for_items = run_in_parallel(
            internal_item_processor.generate_existing_manifests_map,
            [FData(args=[item]) for item in push_items],
            threads=self.target_settings.get("quay_parallelism", 10),
        )

        for item, existing_manifests in zip(push_items, existing_manifests_for_items.values()):
            destination_tags = external_item_processor.generate_repo_dest_tag_map(item)

            for registry, repos in existing_manifests.items():
                for e_repo, _ in repos.items():
                    full_repo = internal_item_processor.reference_processor.replace_repo(e_repo)
                    for d_tag in destination_tags[list(destination_tags.keys())[0]][e_repo]:
                        if (
                            d_tag in existing_manifests[registry][e_repo]
                            and existing_manifests[registry][e_repo][d_tag]
                        ):
                            man_arch_digs = existing_manifests[registry][e_repo][d_tag]
                            if not all_arches:
                                arch_mads = [
                                    m
                                    for m in cast(List[ManifestArchDigest], man_arch_digs)
                                    if m.arch == "amd64"
                                ]
                            else:
                                arch_mads = cast(List[ManifestArchDigest], man_arch_digs)
                            v2list_mads: List[Optional[ManifestArchDigest]] = cast(
                                List[Optional[ManifestArchDigest]],
                                [
                                    m
                                    for m in cast(List[ManifestArchDigest], man_arch_digs)
                                    if m.type_ == QuayClient.MANIFEST_LIST_TYPE
                                ]
                                or [None],
                            )
                            v2s1_mads: List[Optional[ManifestArchDigest]] = cast(
                                List[Optional[ManifestArchDigest]],
                                [m for m in arch_mads if m.type_ == QuayClient.MANIFEST_V2S1_TYPE]
                                or [None],
                            )
                            v2s2_mads: List[Optional[ManifestArchDigest]] = cast(
                                List[Optional[ManifestArchDigest]],
                                [m for m in arch_mads if m.type_ == QuayClient.MANIFEST_V2S2_TYPE]
                                or [None],
                            )
                            for mads in (v2s1_mads, v2s2_mads, v2list_mads):
                                for mad in mads:
                                    if not mad:
                                        continue
                                    image_data = PushDocker.ImageData(
                                        full_repo,
                                        d_tag,
                                        (
                                            mad.digest
                                            if mad.type_ == QuayClient.MANIFEST_LIST_TYPE
                                            else ""
                                        ),
                                        (
                                            mad.digest
                                            if mad.type_ == QuayClient.MANIFEST_V2S2_TYPE
                                            else ""
                                        ),
                                        (
                                            mad.digest
                                            if mad.type_ == QuayClient.MANIFEST_V2S1_TYPE
                                            else ""
                                        ),
                                    )
                                    backup_tags[image_data] = (
                                        json.loads(mad.manifest),
                                        mad.arch,
                                    )
                        else:
                            rollback_tags.append(
                                PushDocker.ImageData(full_repo, d_tag, None, None, None)
                            )
        rollback_tags = sorted(list(set(rollback_tags)))
        return (backup_tags, rollback_tags)

    @log_step("Perform rollback")
    def rollback(self, backup_tags: Dict[ImageData, str], rollback_tags: List[ImageData]) -> None:
        """
        Perform a rollback.

        Args:
            backup_tags ({ImageData: str}):
                Dictionary mapping of ImageData and a manifest before it was overwritten.
            rollback_tags ([ImageData]):
                List of newly added ImageData.
        """
        # restore overwritten tags to their original values
        schema = "{host}/{repo}:{tag}"
        LOG.info("Restoring tags to their original values")
        for image_data, manifest in sorted(backup_tags.items()):
            image_ref = schema.format(host=self.quay_host, repo=image_data.repo, tag=image_data.tag)
            LOG.info("Restoring tag '{0}'".format(image_ref))
            self.dest_quay_client.upload_manifest(manifest, image_ref)

        # delete tags that didn't previously exist
        LOG.info("Removing newly introduced tags")
        for image_data in rollback_tags:
            image_ref = schema.format(host=self.quay_host, repo=image_data.repo, tag=image_data.tag)
            LOG.info("Removing tag '{0}'".format(image_ref))
            try:
                self.dest_quay_api_client.delete_tag(image_data.repo, image_data.tag)
            except requests.exceptions.HTTPError as e:
                # if error occurred before tag was copied, it may not exist
                # deleting a non-existent tag shouldn't be an error
                if e.response.status_code != 404 and e.response.status_code != 401:
                    raise

    def get_outdated_manifests(
        self, backup_tags: Dict[ImageData, str]
    ) -> List[Tuple[str, str, str]]:
        """Return list of existing manifests which are being replaced with new ones.

        Args:
            backup_tags (dict({ImageData: str}): backup tags generated with generate_backup_mapping.
        Returns:
            List of tuples containing digest, tag, repo identifying manifests
        """
        outdated_signatures = []
        for image_data, manifest in backup_tags.items():
            ext_repo = get_external_container_repo_name(image_data.repo.split("/")[1])
            if image_data.v2s2_digest:
                outdated_signatures.append((image_data.v2s2_digest, image_data.tag, ext_repo))
            if image_data.v2s1_digest:
                outdated_signatures.append((image_data.v2s1_digest, image_data.tag, ext_repo))
        return outdated_signatures

    def fetch_missing_push_items_digests(
        self, push_items: List[Any]
    ) -> Dict[str, Dict[str, Dict[str, Dict[str, Tuple[str, Any]]]]]:
        """Fetch digests for media types which weren't originally pushed.

        In order to be able to sign v2s1 for images which were pushed as
        v2s2 or to sign v2s2 for images which were pushed as v2s1
        fetch digests of those missing media types from quay and
        set it to item metadata into 'new_digests' mapping.

        Args:
            push_items(list): List of push items.
            target_settings(dict): Target settings.
        """
        item_processor = item_processor_for_internal_data(
            self.dest_quay_client,
            self.target_settings["quay_host"].rstrip("/"),
            self.dest_registries,
            self.target_settings.get("retry_sleep_time", 5),
            self.target_settings["quay_namespace"],
        )

        new_digests: Dict[str, Dict[str, Dict[str, Dict[str, Tuple[str, Any]]]]] = {}
        for item in push_items:
            missing_media_types = set(
                [QuayClient.MANIFEST_V2S2_TYPE, QuayClient.MANIFEST_V2S1_TYPE]
            ) - set(item.metadata["build"]["extra"]["image"]["media_types"])
            # Always add v2s1 due to possible digest change
            missing_media_types.add(QuayClient.MANIFEST_V2S1_TYPE)
            existing_manifests = item_processor.generate_existing_manifests_map(
                item, only_media_types=list(missing_media_types)
            )
            for reference, repos in existing_manifests.items():
                new_digests.setdefault(reference, {})
                for repo, tags in repos.items():
                    new_digests.setdefault(reference, {}).setdefault(repo, {})
                    for tag, man_arch_digs in tags.items():
                        if not man_arch_digs:
                            continue
                        new_digests[reference][repo].setdefault(tag, {})
                        for mad in man_arch_digs:
                            new_digests[reference][repo][tag][mad.type_] = (
                                mad.digest,
                                item.claims_signing_key,
                            )
        return new_digests

    def sign_new_manifests(self, docker_push_items: List[Any]) -> List[Tuple[str, str, Any]]:
        """Sign newly pushed images with signers enabled in target settings.

        Args:
            docker_push_items(List[PushItem]): list of docker push items.
        Returns:
            List of tuple (reference, digest, key) representing currently signed images.
        """
        current_signatures = []
        to_sign_new_entries = self.fetch_missing_push_items_digests(docker_push_items)
        to_sign_entries = []
        to_sign_entries_internal = []
        for internal_reg, repo_tags in to_sign_new_entries.items():
            for repo, tag_digests in repo_tags.items():
                for tag, digests in tag_digests.items():
                    for type_, digest_key in digests.items():
                        digest, key = digest_key
                        internal_reference = (
                            f"{internal_reg}/"
                            + self.target_settings["quay_namespace"]
                            + "/"
                            + get_internal_container_repo_name(repo)
                            + ":"
                            + tag
                        )
                        for registry in self.dest_registries:
                            pub_reference = f"{registry}/{repo}:{tag}"
                            # add entries in internal format for cosign
                            to_sign_entries_internal.append(
                                SignEntry(
                                    reference=internal_reference,
                                    pub_reference=pub_reference,
                                    repo=repo,
                                    digest=digest,
                                    signing_key=key,
                                    arch="amd64",
                                )
                            )
                            reference = f"{registry}/{repo}:{tag}"
                            to_sign_entries.append(
                                SignEntry(
                                    reference=reference,
                                    pub_reference="",
                                    repo=repo,
                                    digest=digest,
                                    signing_key=key,
                                    arch="amd64",
                                )
                            )
                            current_signatures.append((reference, digest, key))

        for signer in self.target_settings["signing"]:
            if signer["enabled"]:
                signercls = SIGNER_BY_LABEL[signer["label"]]
                signer = signercls(config_file=signer["config_file"], settings=self.target_settings)
                if signercls.pre_push is True:
                    signer.sign_containers(to_sign_entries, self.task_id)
                else:
                    signer.sign_containers(to_sign_entries_internal, self.task_id)

        return current_signatures

    def run(self) -> None:
        """
        Perform the full push-docker workflow.

        The workflow can be summarized as:
        - Filter out push items to only include container image items
        - Check if the destination repos may be pushed to (using Pyxis)
        - Generate backup mapping that will be used for rollback if something goes wrong.
        - Sign container images using RADAS and upload signatures to Pyxis
        - Push container images to their destinations
        - Generate and push container security manifests for pushed images
        - Fetch digests for missing media types of pushed items
        - Sign manifests for missing media types (has to be done after pushing)
        - Filter out push items to only include operator image items
        - Add operator bundles to index images by using IIB
        - Sign index images using RADAS and upload signatures to Pyxis
        - Push the index images to Quay
        - Remove outdated container signatures
        - (in case of failure) Rollback destination repos to the pre-push state
        """
        # TODO: Do we need to manage push item state?
        # Filter out non-docker push items
        docker_push_items = self.get_docker_push_items()
        # Get operator push items (done early so that possible issues are detected)
        operator_push_items = self.get_operator_push_items()
        # Check if we may push to destination repos
        self.check_repos_validity(docker_push_items, self.hub, self.target_settings)
        # Generate resources for rollback in case there are errors during the push
        backup_tags, rollback_tags = self.generate_backup_mapping(
            docker_push_items, all_arches=True
        )
        amd64_backup_tags = {k: bt[0] for k, bt in backup_tags.items() if bt[1] == "amd64"}
        # update backup_tags with also manifest_lists
        amd64_backup_tags.update({k: bt[0] for k, bt in backup_tags.items() if k.v2list_digest})
        all_backup_tags = {k: bt[0] for k, bt in backup_tags.items()}

        existing_index_images = []
        iib_results = None
        successful_iib_results = dict()
        index_stamp = timestamp()
        item_processor = item_processor_for_external_data(
            self.src_quay_client,
            self.dest_registries,
            self.target_settings.get("retry_sleep_time", 5),
        )
        to_sign_entries = []
        current_signatures = []
        to_sign_map = run_in_parallel(
            item_processor.generate_to_sign,
            [FData(args=(item,), kwargs={}) for item in docker_push_items],
        )
        for _to_sign_entries in to_sign_map.values():
            to_sign_entries.extend(_to_sign_entries)

        for sign_entry in to_sign_entries:
            current_signatures.append(
                (sign_entry.reference, sign_entry.digest, sign_entry.signing_key)
            )
        # Sign containers with signers that doesn't need have pushed containers
        # in destination registry
        set_aws_kms_environment_variables(self.target_settings, "cosign_signer")
        for signer in self.target_settings["signing"]:
            if signer["enabled"] and SIGNER_BY_LABEL[signer["label"]].pre_push:
                signercls = SIGNER_BY_LABEL[signer["label"]]
                signer = signercls(config_file=signer["config_file"], settings=self.target_settings)
                signer.sign_containers(to_sign_entries, self.task_id)

        # Push container images
        container_pusher = ContainerImagePusher(docker_push_items, self.target_settings)
        container_pusher.push_container_images()

        # Sign containers with signers which requires pushed containers in destination registry
        to_sign_entries = []
        item_processor = item_processor_for_internal_data(
            self.src_quay_client,
            self.target_settings["quay_host"].rstrip("/"),
            self.dest_registries,
            self.target_settings.get("retry_sleep_time", 5),
            self.target_settings["quay_namespace"],
        )
        to_sign_map = run_in_parallel(
            item_processor.generate_to_sign,
            [
                FData(args=(item,), kwargs={"include_manifest_lists": True})
                for item in docker_push_items
            ],
        )
        for _to_sign_entries in to_sign_map.values():
            to_sign_entries.extend(_to_sign_entries)

        for signer in self.target_settings["signing"]:
            if signer["enabled"] and not SIGNER_BY_LABEL[signer["label"]].pre_push:
                signercls = SIGNER_BY_LABEL[signer["label"]]
                signer = signercls(config_file=signer["config_file"], settings=self.target_settings)
                signer.sign_containers(to_sign_entries, self.task_id)

        self.sign_new_manifests(docker_push_items)

        if self.target_settings.get("push_security_manifests_enabled", False):
            # Generate and push security manifests (if enabled in target settings)
            set_aws_kms_environment_variables(self.target_settings, "security_manifest_signer")
            sec_manifest_pusher = SecurityManifestPusher(docker_push_items, self.target_settings)
            sec_manifest_pusher.push_security_manifests()

        failed = False

        if operator_push_items:
            # Build index images
            operator_pusher = OperatorPusher(
                operator_push_items, self.task_id, self.target_settings
            )
            existing_index_images = operator_pusher.get_existing_index_images(
                self.dest_operator_quay_client
            )
            if operator_pusher.ensure_bundles_present():
                bundles_presence_check_failed = False
                iib_results = operator_pusher.build_index_images()
            else:
                bundles_presence_check_failed = True
                iib_results = {}

            # Sign operator images
            failed_items = [item for item in operator_push_items if item.errors]
            successful_iib_results = dict(
                [(key, val) for key, val in iib_results.items() if val["iib_result"]]
            )

            image_schema = "{host}/{namespace}/{repo}:{tag}"

            for version, iib_details in sorted(successful_iib_results.items()):
                iib_result = iib_details["iib_result"]
                _, iib_namespace, iib_intermediate_repo = parse_index_image(iib_result)
                permanent_index_image = image_schema.format(
                    host=self.target_settings.get("quay_host", "quay.io").rstrip("/"),
                    namespace=iib_namespace,
                    repo=iib_intermediate_repo,
                    tag=iib_result.build_tags[0],
                )
                timestamp_tags = [f"{tag}-{index_stamp}" for tag in iib_details["destination_tags"]]
                current_signatures.extend(
                    _sign_index_image(
                        permanent_index_image,
                        iib_details["destination_tags"] + timestamp_tags,
                        iib_details["signing_keys"],
                        self.task_id,
                        self.target_settings,
                        pre_push=True,
                    )
                )

            # If there are any failed items, skip pushing
            if not failed_items:
                # Push index images to Quay
                operator_pusher.push_index_images(successful_iib_results, index_stamp)

            for version, iib_details in sorted(successful_iib_results.items()):
                iib_result = iib_details["iib_result"]
                _, iib_namespace, iib_intermediate_repo = parse_index_image(iib_result)
                permanent_index_image = image_schema.format(
                    host=self.target_settings.get("quay_host", "quay.io").rstrip("/"),
                    namespace=iib_namespace,
                    repo=iib_intermediate_repo,
                    tag=iib_result.build_tags[0],
                )
                timestamp_tags = [f"{tag}-{index_stamp}" for tag in iib_details["destination_tags"]]
                current_signatures.extend(
                    _sign_index_image(
                        permanent_index_image,
                        iib_details["destination_tags"] + timestamp_tags,
                        iib_details["signing_keys"],
                        self.task_id,
                        self.target_settings,
                        pre_push=False,
                    )
                )

            # Rollback only when all index image builds fails or there are failed items
            # Empty iib_results is not an error and shouldn't fail the push. The only exception is
            # when bundles presence check failed.
            if (
                (iib_results or bundles_presence_check_failed)
                and (not any([x["iib_result"] for x in iib_results.values()]))
                or failed_items
            ):
                if failed_items:
                    LOG.error("There are failed push items. Cannot continue, running rollback.")
                else:
                    LOG.error("Push of all index images failed, running rollback.")
                self.rollback(amd64_backup_tags, rollback_tags)
                sys.exit(1)
            if successful_iib_results != iib_results:
                LOG.error("Push of some index images failed")
                failed = True

        # Remove old signatures
        # run generate backup mapping again to fetch new digests of pushed containers
        backup_tags2, _ = self.generate_backup_mapping(docker_push_items, all_arches=True)
        # if new backup tag has differnet digest, it means it was overwritten during the push
        # and old signature should be removed. If the digest is the same it means, same item
        # was just repushed
        outdated_tags = {}
        backup_tags2_shared = {}

        # Backup tags can contain new tags which were orignally rollback_tags
        # limit the comparision for outdated manifests to original backup_tags only
        for bt2 in backup_tags2.items():
            if (bt2[0].tag, bt2[0].repo) in [(x.tag, x.repo) for x in all_backup_tags.keys()]:
                backup_tags2_shared[bt2[0]] = bt2[1]

        for bt1, bt2 in zip(
            sorted(
                backup_tags.items(),
                key=lambda x: (
                    x[0].tag,
                    x[0].repo,
                    x[0].v2list_digest or "",
                    x[0].v2s2_digest or "",
                    x[0].v2s1_digest or "",
                ),
            ),
            sorted(
                backup_tags2_shared.items(),
                key=lambda x: (
                    x[0].tag,
                    x[0].repo,
                    x[0].v2list_digest or "",
                    x[0].v2s2_digest or "",
                    x[0].v2s1_digest or "",
                ),
            ),
        ):
            list_test = (
                bt1[0].v2list_digest
                and bt2[0].v2list_digest
                and bt1[0].v2list_digest != bt2[0].v2list_digest
            )
            s2_test = (
                bt1[0].v2s2_digest
                and bt2[0].v2s2_digest
                and bt1[0].v2s2_digest != bt2[0].v2s2_digest
            )
            s1_test = (
                bt1[0].v2s1_digest
                and bt2[0].v2s1_digest
                and bt1[0].v2s1_digest != bt2[0].v2s1_digest
            )
            outdated = False
            if list_test:
                d1 = bt1[0].v2list_digest
                d2 = bt2[0].v2list_digest
                name = "manifest list"
                outdated = True
            elif s2_test:
                d1 = bt1[0].v2s2_digest
                d2 = bt2[0].v2s2_digest
                name = "v2s2"
                outdated = True
            elif s1_test:
                d1 = bt1[0].v2s1_digest  # noqa: F841
                d2 = bt2[0].v2s1_digest  # noqa: F841
                name = "v2s1"  # noqa: F841
                outdated = True
            if outdated:
                LOG.debug(
                    f"Marking manifest {bt1[0].tag} as outdated, {name} "
                    f"digests don't match {d1} != {d2}"
                )
                outdated_tags[bt1[0]] = bt1[1]

        outdated_manifests = self.get_outdated_manifests(outdated_tags)
        outdated_manifests.extend(existing_index_images)

        for signer in self.target_settings["signing"]:
            if signer["enabled"] and outdated_manifests:
                signercls = SIGNER_BY_LABEL[signer["label"]]
                signer = signercls(config_file=signer["config_file"], settings=self.target_settings)
                signer.remove_signatures(outdated_manifests, _exclude=current_signatures)

        if failed:
            # Why???
            sys.exit(1)


def mod_entry_point(
    push_items: List[Any], hub: Any, task_id: str, target_name: str, target_settings: Dict[str, Any]
) -> None:
    """Entry point for use in another python code."""
    push = PushDocker(push_items, hub, task_id, target_name, target_settings)
    push.run()



import mock
import pytest

from requests.exceptions import HTTPError

from pubtools._quay.item_processor import (
    ItemProcesor,
    ReferenceProcessorExternal,
    ContentExtractor,
    ReferenceProcessorInternal,
    VirtualPushItem,
)


def test_generate_existing_tags_tolerate_missing(operator_signing_push_item):
    rp = ReferenceProcessorExternal()
    mock_client = mock.MagicMock()
    mock_client.get_repository_tags.side_effect = HTTPError(
        response=mock.MagicMock(status_code=404)
    )
    ip = ItemProcesor(
        source_registry="test-registry.io",
        reference_registries=["dest-registry.io"],
        public_registries=[],
        reference_processor=rp,
        extractor=ContentExtractor(quay_client=mock_client),
    )
    tags = list(ip.generate_existing_tags(operator_signing_push_item))
    assert tags == [("test-registry.io", "repo1", None), ("test-registry.io", "repo2", None)]


def test_generate_existing_tags_no_tolerate_missing(operator_signing_push_item):
    rp = ReferenceProcessorExternal()
    mock_client = mock.MagicMock()
    mock_client.get_repository_tags.side_effect = HTTPError(
        response=mock.MagicMock(status_code=404)
    )
    ip = ItemProcesor(
        source_registry="test-registry.io",
        reference_registries=["dest-registry.io"],
        public_registries=[],
        reference_processor=rp,
        extractor=ContentExtractor(quay_client=mock_client),
    )
    with pytest.raises(HTTPError):
        list(ip.generate_existing_tags(operator_signing_push_item, tolerate_missing=False))


def test_generate_existing_tags_server_error(operator_signing_push_item):
    rp = ReferenceProcessorExternal()
    mock_client = mock.MagicMock()
    mock_client.get_repository_tags.side_effect = HTTPError(
        response=mock.MagicMock(status_code=500)
    )
    ip = ItemProcesor(
        source_registry="test-registry.io",
        reference_registries=["dest-registry.io"],
        public_registries=[],
        reference_processor=rp,
        extractor=ContentExtractor(quay_client=mock_client),
    )
    with pytest.raises(HTTPError):
        list(ip.generate_existing_tags(operator_signing_push_item, tolerate_missing=False))


def test_generate_existing_tags_server_error_tolerate_missing(operator_signing_push_item):
    rp = ReferenceProcessorExternal()
    mock_client = mock.MagicMock()
    mock_client.get_repository_tags.side_effect = HTTPError(
        response=mock.MagicMock(status_code=500)
    )
    ip = ItemProcesor(
        source_registry="test-registry.io",
        reference_registries=["dest-registry.io"],
        public_registries=[],
        reference_processor=rp,
        extractor=ContentExtractor(quay_client=mock_client),
    )
    with pytest.raises(HTTPError):
        list(ip.generate_existing_tags(operator_signing_push_item, tolerate_missing=True))


def test_reference_processor_internal():
    assert (
        ReferenceProcessorInternal(quay_namespace="ns").full_reference(
            "registry", "namespace/repo", None
        )
        == "registry/ns/namespace----repo"
    )


def test_reference_processor_internal_no_slash():
    assert (
        ReferenceProcessorInternal(quay_namespace="ns").full_reference(
            "registry", "noslash-repo", None
        )
        == "registry/ns/noslash-repo"
    )


def test_reference_processor_internal_no_slash_tag():
    assert (
        ReferenceProcessorInternal(quay_namespace="ns").full_reference(
            "registry", "noslash-repo", "tag"
        )
        == "registry/ns/noslash-repo:tag"
    )


def test_reference_processor_internal_invalid_repo():
    with pytest.raises(ValueError):
        assert ReferenceProcessorInternal(quay_namespace="ns").full_reference(
            "registry", "namespace/ns/repo", None
        )


def test_reference_processor_internal_tag():
    assert (
        ReferenceProcessorInternal(quay_namespace="ns").full_reference(
            "registry", "namespace/repo", "tag"
        )
        == "registry/ns/namespace----repo:tag"
    )


def test_reference_processor_replace_repo():
    assert (
        ReferenceProcessorInternal(quay_namespace="ns").replace_repo("namespace/repo")
        == "ns/namespace----repo"
    )


def test_reference_processor_replace_repo_no_slash():
    assert (
        ReferenceProcessorInternal(quay_namespace="ns").replace_repo("namespace-repo")
        == "ns/namespace-repo"
    )


def test_reference_processor_replace_repo_error():
    with pytest.raises(ValueError):
        assert ReferenceProcessorInternal(quay_namespace="ns").replace_repo("namespace/ns/repo")


def test_invalid_virtual_push_item():
    with pytest.raises(ValueError):
        VirtualPushItem(repos=[], metadata={})


def test_reference_processor_external_tag():
    assert (
        ReferenceProcessorExternal().full_reference("registry", "namespace/repo", "tag")
        == "registry/namespace/repo:tag"
    )


def test_reference_processor_external_repo():
    assert (
        ReferenceProcessorExternal().full_reference("registry", "namespace/repo", None)
        == "registry/namespace/repo"
    )


def test_generate_existing_manifest_map_tolerate_429(operator_signing_push_item):
    rp = ReferenceProcessorExternal()
    mock_client = mock.MagicMock()
    mock_client.get_manifest.side_effect = HTTPError(response=mock.MagicMock(status_code=429))
    ip = ItemProcesor(
        source_registry="test-registry.io",
        reference_registries=["dest-registry.io"],
        public_registries=[],
        reference_processor=rp,
        extractor=ContentExtractor(quay_client=mock_client, sleep_time=0),
    )
    assert ip.generate_existing_manifests_map(operator_signing_push_item) == {
        "test-registry.io": {
            "repo1": {"latest-test-tag": None, "1.0": None},
            "repo2": {"tag2": None},
        }
    }



from dataclasses import dataclass
import hashlib
import logging
from typing import List, Dict, Any, Optional, Tuple, cast, Callable, Union
from typing_extensions import TypeAlias
import requests
import time
import json

from .quay_client import QuayClient
from .exceptions import ManifestTypeError
from .utils.misc import run_in_parallel, FData

LOG = logging.getLogger("pubtools.quay")


@dataclass
class VirtualPushItem:
    """Virtual push item used for operations which doens't take push item on the input."""

    metadata: Dict[str, Any]
    repos: Dict[str, Any]

    def __post_init__(self) -> None:
        """Post init validation of the instance."""
        if "tags" not in self.metadata:
            raise ValueError("VirtualPushItem is missing 'tags' in metadata")


PushItem: TypeAlias = VirtualPushItem


@dataclass
class SignEntry:
    """Data structure to hold signing related information.

    Args:
        signing_key (str): Signing key.
        repo (str): Repo reference in format <registry>/<repo>
        reference (str): Reference in format <registry>/<repo>:<tag>
        pub_reference (str): Public repo reference in format <registry>/repo
        digest (str): Digest of the manifest.
        arch (str): Architecture of the manifest.
    """

    repo: str
    reference: str
    pub_reference: str
    digest: str
    signing_key: str
    arch: str


@dataclass(frozen=True)
class ManifestArchDigest:
    """Data structure to hold information about container manifest."""

    manifest: str
    digest: str
    arch: str
    type_: str


class ContentExtractor:
    """Class is used to extract specific content from container registry based on provided input."""

    _MEDIA_TYPES_PRIORITY = {
        QuayClient.MANIFEST_LIST_TYPE: 30,
        QuayClient.MANIFEST_V2S2_TYPE: 20,
        QuayClient.MANIFEST_V2S1_TYPE: 10,
    }

    def __init__(
        self,
        quay_client: QuayClient,
        sleep_time: int = 5,
        timeout: int = 10,
        poll_rate: int = 5,
        full_extract: bool = False,
    ):
        """Initialize the class.

        Args:
            quay_client (QuayClient): Quay client used to communicate with container registry.
            sleep_time (int): Time to sleep between retries.
            timeout (int): Timeout for HTTP requests.
            poll_rate (int): Poll rate for HTTP requests.
        """
        self.quay_client = quay_client
        self.sleep_time: int = sleep_time
        self.timeout: int = timeout
        self.poll_rate: int = poll_rate
        self.full_extract = full_extract

    def _extract_ml_manifest_full(
        self, image_ref: str, _manifest_list: str, mtype: str, ret_headers: Dict[str, Any]
    ) -> List[ManifestArchDigest]:
        """Extract manifests from manifest list.

        Args:
            image_ref (str): Image reference in format <registry>/<repo>:<tag>
            _manifest_list (str): Manifest list in JSON format.
            mtype (str): NOT USED
        Returns:
            list: List of ManifestArchDigest objects.
        """
        mads = []
        manifest = json.loads(_manifest_list)
        mads.append(
            ManifestArchDigest(
                manifest=_manifest_list,
                digest=ret_headers.get("docker-content-digest", ""),
                arch="",
                type_=QuayClient.MANIFEST_LIST_TYPE,
            )
        )
        for arch_manifest in manifest["manifests"]:
            manifest = self.quay_client.get_manifest(
                f"{image_ref.rsplit(':')[0]}@{arch_manifest['digest']}",
                media_type=QuayClient.MANIFEST_V2S2_TYPE,
                raw=True,
            )
            mads.append(
                ManifestArchDigest(
                    manifest=manifest,
                    digest=arch_manifest["digest"],
                    arch=arch_manifest["platform"]["architecture"],
                    type_=QuayClient.MANIFEST_V2S2_TYPE,
                )
            )
        return mads

    def _extract_ml_manifest(
        self, image_ref: str, _manifest_list: str, mtype: str, ret_headers: Dict[str, Any]
    ) -> List[ManifestArchDigest]:
        """Extract manifests from manifest list.

        Args:
            image_ref (str): Image reference in format <registry>/<repo>:<tag>
            _manifest_list (str): Manifest list in JSON format.
            mtype (str): NOT USED
        Returns:
            list: List of ManifestArchDigest objects.
        """
        mads = []
        mads.append(
            ManifestArchDigest(
                manifest=_manifest_list,
                digest=ret_headers.get("docker-content-digest", ""),
                arch="",
                type_=QuayClient.MANIFEST_LIST_TYPE,
            )
        )
        return mads

    def _extract_manifest(
        self, unused: str, manifest: str, mtype: str, ret_headers: Dict[str, Any]
    ) -> ManifestArchDigest:
        """Calculate information from given manifest.

        Args:
            unused (str): This parameter is unused for this specific method.
            It's definied like this to simply  keep same signature as similar
            method `_extract_ml_manifest`
            manifest (str): Manifest
            mtype (str): Media type of the manifest.
        Returns:
            ManifestArchDigest: ManifestArchDigest object.
        """
        hasher = hashlib.sha256()
        hasher.update(manifest.encode("utf-8"))
        digest = f"sha256:{hasher.hexdigest()}"
        return ManifestArchDigest(manifest=manifest, digest=digest, arch="amd64", type_=mtype)

    _MEDIA_TYPES_PROCESS: Dict[
        str,
        Callable[
            [Any, str, str, str, Dict[str, Any]],
            Union[ManifestArchDigest, List[ManifestArchDigest]],
        ],
    ] = {
        QuayClient.MANIFEST_LIST_TYPE: _extract_ml_manifest,
        QuayClient.MANIFEST_V2S2_TYPE: _extract_manifest,
        QuayClient.MANIFEST_V2S1_TYPE: _extract_manifest,
    }

    _MEDIA_TYPES_PROCESS_FULL: Dict[
        str,
        Callable[
            [Any, str, str, str, Dict[str, Any]],
            Union[ManifestArchDigest, List[ManifestArchDigest]],
        ],
    ] = {
        QuayClient.MANIFEST_LIST_TYPE: _extract_ml_manifest_full,
        QuayClient.MANIFEST_V2S2_TYPE: _extract_manifest,
        QuayClient.MANIFEST_V2S1_TYPE: _extract_manifest,
    }

    def extract_manifests(
        self, image_ref: str, media_types: Optional[List[str]], tolerate_missing: bool = True
    ) -> List[ManifestArchDigest]:
        """Extract manifests from container registry.

        Method fetches manifests for all provided media types. If HTTPErrors is raised when
        fetching for manifest and status code is 404 or 401, it's retried as there can be delay
        in registry if manifest was just pushed to it.
        When manifest is not found for given media type, it's not included in the result.
        media_types can be also empty and it that case all container media types are used

        Args:
            image_ref (str): Image reference in format <registry>/<repo>:<tag>
            media_types (list): List of media types to be extracted.
            tolerate_missing (bool): If True, missing manifests are tolerated and
            empty list is returned.

        Returns:
            list: List of ManifestArchDigest objects.
        """
        mtypes = (
            sorted(media_types or [], key=lambda x: self._MEDIA_TYPES_PRIORITY[x], reverse=True)
            or self._MEDIA_TYPES_PRIORITY.keys()
        )
        if self.full_extract:
            MEDIA_TYPES_PROCESS = self._MEDIA_TYPES_PROCESS_FULL
        else:
            MEDIA_TYPES_PROCESS = self._MEDIA_TYPES_PROCESS
        results = []
        for mtype in mtypes:
            ret_headers: Dict[str, Any] = {}
            for i in range(self.timeout // self.poll_rate):
                try:
                    ret = cast(
                        Tuple[str, Dict[str, Any]],
                        self.quay_client.get_manifest(
                            image_ref, media_type=mtype, raw=True, return_headers=True
                        ),
                    )
                    if ret:
                        manifest, ret_headers = ret
                    else:
                        manifest = None
                except requests.exceptions.HTTPError as e:
                    # When robot account is used, 401 may be returned instead of 404
                    if (
                        e.response.status_code == 404 or e.response.status_code == 401
                    ) and tolerate_missing:
                        manifest = None
                    # tolerate too many requests from client
                    elif e.response.status_code == 429:
                        manifest = None
                    else:
                        raise
                except ManifestTypeError:
                    manifest = None
                    break
                else:
                    break
                time.sleep(self.sleep_time)
            if not manifest:
                continue
            else:
                mret: Union[ManifestArchDigest, List[ManifestArchDigest]] = MEDIA_TYPES_PROCESS[
                    mtype
                ](self, image_ref, manifest, mtype, ret_headers)
                if isinstance(mret, list):
                    results.extend(mret)
                else:
                    results.append(mret)
        seen = set()
        return [x for x in results if all([x not in seen, not (seen.add(x))])]  # type: ignore

    def extract_tags(self, repo_ref: str, tolerate_missing: bool = True) -> List[str]:
        """Fetch list of tags for given repo reference.

        Args:
            repo_ref (str): Repo reference in format <registry>/<repo>
            tolerate_missing (bool): If True, missing repo is tolerated and empty list
            is returned.
        Returns:
            list: List of tags.
        """
        try:
            repo_tags = cast(Dict[str, List[str]], self.quay_client.get_repository_tags(repo_ref))
        except requests.exceptions.HTTPError as e:
            # When robot account is used, 401 is returned instead of 404
            if tolerate_missing and e.response.status_code == 404 or e.response.status_code == 401:
                repo_tags = {"tags": []}
            else:
                raise
        return repo_tags["tags"]


class ReferenceProcessorExternal:
    """Class is used to produce full image reference or repo reference from input."""

    def full_reference(self, registry: str, repo: str, tag: Optional[str] = None) -> str:
        """Produce full image reference from input.

        Args:
            registry (str): Registry where image is located.
            repo (str): Repository name.
            tag (Optional[str]): Tag name, if not set only repo reference is returned
        Returns:
            str: full image reference.
        """
        if tag:
            return f"{registry}/{repo}:{tag}"
        else:
            return f"{registry}/{repo}"

    def replace_repo(self, repo: str) -> str:
        """Return repo unmodified.

        External reference processor does not modify repo.

        Args:
            repo (str): Repository in format <namespace>/<product>
        Returns:
            str: Repository
        """
        return repo


@dataclass
class ReferenceProcessorInternal:
    """Class is used to produce full internal image reference from input."""

    INTERNAL_DELIMITER = "----"
    quay_namespace: str

    def full_reference(self, registry: str, repo: str, tag: Optional[str] = None) -> str:
        """Produce full internal image reference from input.

        Args:
            registry (str): Registry where image is located.
            repo (str): Repository name.
            tag (Optional[str]): Tag name.
        Returns:
            str: full image reference.
        """
        if repo.count("/") == 0:
            if tag:
                return f"{registry}/{self.quay_namespace}/{repo}:{tag}"
            else:
                return f"{registry}/{self.quay_namespace}/{repo}"
        if repo.count("/") > 1 or repo[0] == "/" or repo[-1] == "/":
            raise ValueError(
                "Input repository containing a delimeter should "
                "have the format '<namespace>/<product>' or '<repository>'",
                repo,
            )
        replaced = repo.replace("/", self.INTERNAL_DELIMITER)
        if tag:
            return f"{registry}/{self.quay_namespace}/{replaced}:{tag}"
        else:
            return f"{registry}/{self.quay_namespace}/{replaced}"

    def replace_repo(self, repo: str) -> str:
        """Convert repo to internal format.

        Args:
            repo (str): Repository in format <namespace>/<product>
        Returns:
            str: Repository in format <quay_namespace>/<namespace>----<product>
        """
        if repo.count("/") == 0:
            return f"{self.quay_namespace}/{repo}"
        if repo.count("/") > 1 or repo[0] == "/" or repo[-1] == "/":
            raise ValueError(
                "Input repository containing a delimeter should "
                "have the format '<namespace>/<product>' or '<repository>'",
                repo,
            )
        replaced = repo.replace("/", self.INTERNAL_DELIMITER)
        return f"{self.quay_namespace}/{replaced}"


@dataclass
class ItemProcesor:
    """Class is used to process push item and extract various data from it.

    Args:
        extractor (ContentExtractor): Content extractor which is used when container metadata for
        processed push item is needed.
        reference_processor (ReferenceProcessor): Reference processor to produce reference from
        provided push item. Based on used reference processor it can produce reference to internal
        or external push item.
        reference_registries (List[str]): List of destination registries where push item should be
        available when pushed. Can be empty if there's no need to generate destination data.
        source_registry (str): Source registry where source container image is located.
        Can be empty if there's no need to generate source data.
    """

    extractor: ContentExtractor
    reference_processor: Union[ReferenceProcessorExternal, ReferenceProcessorInternal]
    reference_registries: List[str]
    source_registry: Optional[str]
    public_registries: List[str]

    INTERNAL_DELIMITER = "----"

    def _generate_dest_repo(self, item: PushItem) -> List[Tuple[str, str]]:
        """Return list of (<dest-registry>, <dest-repository>) tuples for given push item.

        Args:
            item (PushItem): Push item.
        Returns:
            list: List of tuples containing registry and repository.
        """
        dest_registry_repos = []
        for registry in self.reference_registries:
            for repo in item.metadata["tags"].keys():
                dest_registry_repos.append((registry, repo))
        return dest_registry_repos

    def _generate_src_repo(self, item: PushItem) -> List[str]:
        """Return list of source repos for given push item.

        Returns:
            list: List of source repositories.
        """
        return list(item.repos.keys())

    def _generate_src_repo_tag(self, item: PushItem) -> List[Tuple[str, str]]:
        """Return list of tuples of (<source-repo>, <tag>).

        Args:
            item (PushItem): Push item.
        Returns:
            list: List of tuples containing source repository and tag.
        """
        src_repo_tag = []
        for repo, tags in item.metadata["tags"].items():
            for tag in tags:
                src_repo_tag.append((repo, tag))
        return src_repo_tag

    def generate_repo_dest_tags(self, item: PushItem) -> List[Tuple[str, str, str]]:
        """Return list of destination repositories and tags.

        Args:
            item (PushItem): Push item.
        Returns:
            list: List of tuples containing registry, repository and tag.
        """
        ret = []
        for registry, repo in self._generate_dest_repo(item):
            for repo, tags in item.metadata["tags"].items():
                for tag in tags:
                    ret.append((registry, repo, tag))
        return ret

    def generate_repo_untags(self, item: PushItem) -> List[Tuple[str, str]]:
        """Return list of repositories and tags which are destined to be untag.

        Args:
            item (PushItem): Push item.
        Returns:
            list: List of tuples containing repository and tag.
        """
        ret = []
        for repo in self._generate_src_repo(item):
            for tag in item.metadata["remove_tags"]:
                ret.append((repo, tag))
        return ret

    def generate_repo_dest_tag_map(self, item: PushItem) -> Dict[str, Dict[str, List[str]]]:
        """Return map of destination repositories and tags.

        Args:
            item (PushItem): Push item.
        Returns:
            dict: Dict of {registry: {repo: [tags]}}
        """
        ret: dict[str, Dict[str, list[str]]] = {}
        for registry, repo in self._generate_dest_repo(item):
            for repo, tags in item.metadata["tags"].items():
                for tag in tags:
                    ret.setdefault(registry, {}).setdefault(repo, []).append(tag)
        return ret

    def generate_to_sign(self, item: Any, include_manifest_lists: bool = False) -> List[SignEntry]:
        """Generate list of images to sign.

        Args:
            item (PushItem): Push item.
            sign_only_arches(List[str]): List of architectures to sign.
            If empty, all sign architectures.
        Returns:
            list: List of dictionaries containing reference, digest, repository and architecture.
        """
        to_sign = []
        media_types = (
            item.metadata.get("build", {}).get("extra", {}).get("image", {}).get("media_types", [])
        )

        full_extract_orig = self.extractor.full_extract
        self.extractor.full_extract = True
        man_arch_digs = self.extractor.extract_manifests(item.metadata["pull_url"], media_types)
        self.extractor.full_extract = full_extract_orig

        for registry, repo, tag in self.generate_repo_dest_tags(item):
            reference = self.reference_processor.full_reference(registry, repo, tag)
            for mad in man_arch_digs:
                if mad.type_ == QuayClient.MANIFEST_LIST_TYPE and not include_manifest_lists:
                    continue
                public_registries = self.public_registries if registry == "quay.io" else [registry]
                for public_registry in public_registries:
                    to_sign.append(
                        SignEntry(
                            repo=repo,
                            reference=reference,
                            pub_reference=f"{public_registry}/{repo}:{tag}",
                            digest=mad.digest,
                            arch=mad.arch,
                            signing_key=item.claims_signing_key,
                        )
                    )
        return to_sign

    def generate_to_unsign(self, item: PushItem) -> List[Dict[str, Any]]:
        """Generate list of images to unsign.

        Args:
            item (PushItem): Push item.
        Returns:
            list: List of dictionaries containing reference, digest, repository and architecture.
        """
        to_unsign = []
        media_types = (
            item.metadata.get("build", {}).get("extra", {}).get("image", {}).get("media_types", [])
        )
        references = []
        for repo, tag in self.generate_repo_untags(item):
            references.append(
                self.reference_processor.full_reference(cast(str, self.source_registry), repo, tag)
            )

        full_extract_orig = self.extractor.full_extract
        self.extractor.full_extract = True
        man_arch_digs_map = run_in_parallel(
            self.extractor.extract_manifests, [FData(args=(ref, media_types)) for ref in references]
        )
        self.extractor.full_extract = full_extract_orig

        for ref_index, man_arch_digs in man_arch_digs_map.items():
            for mad in man_arch_digs:
                to_unsign.append(
                    {
                        "reference": references[ref_index],
                        "digest": mad.digest,
                        "repo": repo,
                        "arch": mad.arch,
                    }
                )
        return to_unsign

    def generate_existing_tags(
        self, item: PushItem, tolerate_missing: bool = True
    ) -> List[Tuple[str, str, Union[str, None]]]:
        """Generate list of existing tags for given push item.

        Args:
            item (PushItem): Push item.
            tolerate_missing (bool): If True, tolerate missing tags.
        Returns:
            list: List of tuples containing registry, repository and tag.
        """
        existing_tag_entries: List[Tuple[str, str, Union[str, None]]] = []
        for repo in self._generate_src_repo(item):
            ref_repo = self.reference_processor.replace_repo(repo)
            tags = self.extractor.extract_tags(ref_repo, tolerate_missing=tolerate_missing)
            for tag in tags:
                existing_tag_entries.append((cast(str, self.source_registry), repo, tag))
            if not tags:
                existing_tag_entries.append((cast(str, self.source_registry), repo, None))
        return existing_tag_entries

    def _generate_existing_manifests(
        self, item: PushItem, only_media_types: Union[List[str], None] = None
    ) -> List[Tuple[str, str, Optional[ManifestArchDigest]]]:
        """Generate list of existing manifests data for given push item.

        Args:
            item (PushItem): Push item.
            only_media_types (list): Restrict to specific media types.
        Returns:
            list: List of tuples containing repository, tag and ManifestArchDigest.
        """
        existing_manifests: List[Tuple[str, str, Union[ManifestArchDigest, None]]] = []
        if not only_media_types:
            media_types = [
                QuayClient.MANIFEST_LIST_TYPE,
                QuayClient.MANIFEST_V2S2_TYPE,
                QuayClient.MANIFEST_V2S1_TYPE,
            ]
        else:
            media_types = only_media_types

        references = []
        for repo, tag in self._generate_src_repo_tag(item):
            full_ref = self.reference_processor.full_reference(
                cast(str, self.source_registry), repo, tag=tag
            )
            references.append((full_ref, repo, tag))

        man_arch_digs_map = run_in_parallel(
            self.extractor.extract_manifests,
            [FData(args=(ref_meta[0], media_types)) for ref_meta in references],
        )
        for ref_index, man_arch_digs in man_arch_digs_map.items():
            reference, repo, tag = references[ref_index]
            for mad in man_arch_digs:
                if (repo, tag, mad) not in existing_manifests:
                    existing_manifests.append((repo, tag, mad))
            if not man_arch_digs:
                # If no manifest found, set tag to None
                if (repo, tag, None) not in existing_manifests:
                    existing_manifests.append((repo, tag, None))
        return existing_manifests

    def generate_existing_manifests_map(
        self, item: PushItem, only_media_types: Union[List[str], None] = None
    ) -> Dict[str, Dict[str, Dict[str, Union[List[ManifestArchDigest], None]]]]:
        """Generate existing manifests map for given push item.

        Args:
            item (PushItem): Push item.
            only_media_types (list): List of media types to check.
        Returns:
            dict: Dict of {registry: {repo: {tag: [<ManifestArchDigest>]}}}
        """
        mapping_existing: Dict[str, Dict[str, dict[str, Union[List[ManifestArchDigest], None]]]] = (
            {}
        )
        for repo, tag, mad in self._generate_existing_manifests(
            item, only_media_types=only_media_types
        ):
            if mad:
                cast(
                    List[ManifestArchDigest],
                    mapping_existing.setdefault(cast(str, self.source_registry), {})
                    .setdefault(repo, {})
                    .setdefault(tag, []),
                ).append(mad)
            else:
                mad = cast(None, mad)
                mapping_existing.setdefault(cast(str, self.source_registry), {}).setdefault(
                    repo, {}
                ).setdefault(tag, mad)
        return mapping_existing

    def generate_existing_manifests_metadata(
        self, item: PushItem, only_media_types: Union[List[str], None] = None
    ) -> List[Tuple[str, str, Union[ManifestArchDigest, None]]]:
        """Generate list of existing manifests for given push item.

        Args:
            item (PushItem): Push item.
            only_media_types (list): Restrict to specific media types.
        Returns:
            list: List of tuples containing repository, tag and ManifestArchDigest.
        """
        existing_manifests = []
        for repo, tag, mad in self._generate_existing_manifests(
            item, only_media_types=only_media_types
        ):
            existing_manifests.append((repo, tag, mad))
        return existing_manifests

    def generate_all_existing_manifests_metadata(
        self, item: PushItem
    ) -> List[Tuple[str, str, Union[ManifestArchDigest, None]]]:
        """Return manifests for all existing tags in all repositories for given push item.

        Args:
            item (PushItem): Push item.
        Returns:
            list: List of tuples containing repository, tag and ManifestArchDigest.
        """
        repo_tags_map: dict[str, List[Union[str, None]]] = {}
        for registry, repo, tag in self.generate_existing_tags(item):
            repo_tags_map.setdefault(repo, []).append(tag)
        item2 = VirtualPushItem(
            metadata={"tags": repo_tags_map},
            repos={repo: [] for repo in repo_tags_map.keys()},
        )
        return self.generate_existing_manifests_metadata(item2)


def item_processor_for_external_data(
    quay_client: QuayClient, external_registries: List[str], retry_sleep_time: int
) -> ItemProcesor:
    """Get instance of item processor configured to produce destination data.

    Args:
        quay_client (QuayClient): Quay client.
        external_registries (list): List of external registries.
        retry_sleep_time (int): sleep time bewteen retries for fetching data from registry.
    Returns:
        ItemProcessor: Instance of item processor.
    """
    extractor = ContentExtractor(quay_client=quay_client, sleep_time=retry_sleep_time)
    return ItemProcesor(
        extractor=extractor,
        reference_processor=ReferenceProcessorExternal(),
        reference_registries=external_registries,
        source_registry=None,
        public_registries=external_registries,
    )


def item_processor_for_internal_data(
    quay_client: QuayClient,
    internal_registry: str,
    external_registries: List[str],
    retry_sleep_time: int,
    internal_namespace: str,
) -> ItemProcesor:
    """Get instance of item processor configured to produce internal data.

    Args:
        quay_client (QuayClient): Quay client.
        internal_registry (str): Docker registry where containers are stored
        external registries (str): List of external registries used for container identity.
        retry_sleep_time (int): sleep time bewteen retries for fetching data from registry.
        internal_namespace (str): Namespace of internal organization in the registry.
    Returns:
        ItemProcessor: Instance of item processor.
    """
    extractor = ContentExtractor(quay_client=quay_client, sleep_time=retry_sleep_time)
    reference_processor = ReferenceProcessorInternal(internal_namespace)
    return ItemProcesor(
        extractor=extractor,
        reference_processor=reference_processor,
        reference_registries=["quay.io"],
        source_registry=internal_registry,
        public_registries=external_registries,
    )



import mock
import pytest

from pubtools._quay import quay_session


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_init_docker_api(mock_session):
    mocked_session = mock.MagicMock()
    mocked_session.headers = {}
    mock_session.return_value = mocked_session

    session = quay_session.QuaySession(api="docker")

    assert session.hostname == "quay.io"
    assert session.session.verify is False
    assert session.session.headers["Host"] == "quay.io"
    assert session.api == "docker"
    assert session._api_url("some-endpoint") == "https://quay.io/v2/some-endpoint"


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_init_quay_api(mock_session):
    mocked_session = mock.MagicMock()
    mocked_session.headers = {}
    mock_session.return_value = mocked_session

    session = quay_session.QuaySession(api="quay")

    assert session.hostname == "quay.io"
    assert session.session.verify is False
    assert session.session.headers["Host"] == "quay.io"
    assert session.api == "quay"
    assert session._api_url("some-endpoint") == "https://quay.io/api/v1/some-endpoint"


def test_init_unsupported_api():
    with pytest.raises(ValueError, match="Unknown API type.*"):
        quay_session.QuaySession(api="unsupported")


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_api_url(mock_session):
    session1 = quay_session.QuaySession()
    result1 = session1._api_url("some/endpoint")
    assert result1 == "https://quay.io/v2/some/endpoint"

    session2 = quay_session.QuaySession("http://registry.com")
    result2 = session2._api_url("other/endpoint")
    assert result2 == "http://registry.com/v2/other/endpoint"


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_set_token(mock_session):
    mocked_session = mock.MagicMock()
    mocked_session.headers = {}
    mock_session.return_value = mocked_session

    session = quay_session.QuaySession()
    session.set_auth_token("some-token")

    assert session.session.headers["Authorization"] == "Bearer some-token"


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_get(mock_session):
    mocked_session = mock.MagicMock()
    mocked_session.headers = {}
    mock_session.return_value = mocked_session
    session = quay_session.QuaySession()

    kwargs = {"headers": {"Accept": "application/json"}}
    session.get("get/data/1", **kwargs)
    kwargs["timeout"] = 10
    mocked_session.get.assert_called_with("https://quay.io/v2/get/data/1", **kwargs)


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_post(mock_session):
    mocked_session = mock.MagicMock()
    mocked_session.headers = {}
    mock_session.return_value = mocked_session
    session = quay_session.QuaySession()

    kwargs = {"headers": {"Accept": "application/json"}, "data": "some data"}
    session.post("post/data/2", **kwargs)
    kwargs["timeout"] = 10
    mocked_session.post.assert_called_with("https://quay.io/v2/post/data/2", **kwargs)


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_put(mock_session):
    mocked_session = mock.MagicMock()
    mocked_session.headers = {}
    mock_session.return_value = mocked_session
    session = quay_session.QuaySession()

    kwargs = {"data": "new data"}
    session.put("put/data/3", **kwargs)
    kwargs["timeout"] = 10
    mocked_session.put.assert_called_with("https://quay.io/v2/put/data/3", **kwargs)


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_delete(mock_session):
    mocked_session = mock.MagicMock()
    mocked_session.headers = {}
    mock_session.return_value = mocked_session
    session = quay_session.QuaySession()

    kwargs = {"data": "old data"}
    session.delete("delete/data/4", **kwargs)
    kwargs["timeout"] = 10
    mocked_session.delete.assert_called_with("https://quay.io/v2/delete/data/4", **kwargs)


@mock.patch("pubtools._quay.quay_session.requests.Session")
def test_request(mock_session):
    mocked_session = mock.MagicMock()
    mocked_session.headers = {}
    mock_session.return_value = mocked_session
    session = quay_session.QuaySession()

    kwargs = {"headers": {"Accept": "application/json"}, "data": "some data"}
    session.request("POST", "post/data/2", **kwargs)
    kwargs["timeout"] = 10
    mocked_session.request.assert_called_with("POST", "https://quay.io/v2/post/data/2", **kwargs)



from typing_extensions import TypedDict
from typing import List


Platform = TypedDict(
    "Platform",
    {
        "architecture": str,
        "os": str,
        "os.version": str,
        "os.features": List[str],
        "variant": str,
        "features": List[str],
    },
)


class Manifest(TypedDict):
    """Typed dict use dto store manifest data."""

    mediaType: str
    size: int
    digest: str
    platform: Platform


class ManifestList(TypedDict):
    """Typed dict use dto store manifest list data."""

    schemaVersion: int
    mediaType: str
    manifests: List[Manifest]
    platform: Platform
